__all__ = ('render_exception_into', 'render_frames_into',)

from ..cause_group import CauseGroup

from .exception_proxy import ExceptionProxyRich
from .frame_grouping import group_frames
from .frame_ignoring import should_keep_frame
from .frame_proxy import convert_frames_to_frame_proxies, populate_frame_proxies
from .rendering import render_frame_groups_into, add_trace_title_into, render_exception_proxy_into


def render_frames_into(frames, extend = None, *, filter = None, highlighter = None):
    """
    Renders the given frames into a list of strings.
    
    Parameters
    ----------
    frames : `list<FrameProxyBase | FrameType | TraceBack>`
        The frames to render.
    
    extend : `None`, `list<str>` = `None`, Optional
        Whether the frames should be rendered into an already existing list.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Stores how the output should be highlighted.
    
    Returns
    -------
    extend : `list<str>`
        The rendered frames as a `list` of it's string parts.
    """
    frames = convert_frames_to_frame_proxies(frames)
    populate_frame_proxies(frames)
    
    if extend is None:
        extend = []
    
    frames = [frame for frame in frames if should_keep_frame(frame, filter = filter)]
    frame_groups = group_frames(frames)
    
    render_frame_groups_into(frame_groups, highlighter, extend)
    
    return extend


REASON_TYPE_NONE = 0
REASON_TYPE_CAUSE = 1
REASON_TYPE_CONTEXT = 2
REASON_TYPE_CAUSE_GROUP = 3


def render_exception_into(exception, extend = None, *, filter = None, highlighter = None):
    """
    Renders the given exception's frames into a list of strings.
    
    Parameters
    ----------
    exception : `None | BaseException`
        The exception to render.
    
    extend : `None`, `list<str>` = `None`, Optional
        Whether the frames should be rendered into an already existing list.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Stores how the output should be highlighted.
    
    Returns
    -------
    extend : `list<str>`
        The rendered frames as a `list` of it's string parts.
    """
    if extend is None:
        extend = []
    
    exceptions = []
    reason_type = REASON_TYPE_NONE
    while True:
        exceptions.append((exception, reason_type))
        
        cause_exception = exception.__cause__
        if (cause_exception is not None):
            exception = cause_exception
            if isinstance(exception, CauseGroup):
                reason_type = REASON_TYPE_CAUSE_GROUP
            else:
                reason_type = REASON_TYPE_CAUSE
            continue
        
        if not exception.__suppress_context__:
            context_exception = exception.__context__
            if (context_exception is not None):
                exception = context_exception
                reason_type = REASON_TYPE_CONTEXT
                continue
        
        # no other cases
        break
    
    
    for exception, reason_type in reversed(exceptions):
        if reason_type == REASON_TYPE_CAUSE_GROUP:
            extend = add_trace_title_into(
                f'The following {len(exception)} exceptions where the reason of the exception following them:',
                highlighter,
                extend,
            )
            extend.append('\n\n')
            
            for index, cause in enumerate(exception, 1):
                extend = add_trace_title_into(f'[Exception {index}] ', highlighter, extend)
                extend = render_exception_into(cause, extend = extend, filter = filter, highlighter = highlighter)
                
                if index != len(exception):
                    extend.append('\n')
        
        else:
            extend = add_trace_title_into('Traceback (most recent call last):', highlighter, extend)
            extend.append('\n')
            
            exception_proxy = ExceptionProxyRich(exception)
            populate_frame_proxies([*exception_proxy.iter_frames_no_repeat()])
            exception_proxy.drop_ignored_frames(filter = filter)
            extend = render_exception_proxy_into(exception_proxy, highlighter, extend)
            
            if reason_type == REASON_TYPE_NONE:
                break
        
        if reason_type == REASON_TYPE_CAUSE:
            title = 'The above exception was the direct cause of the following exception:'
        
        elif reason_type == REASON_TYPE_CONTEXT:
            title = 'During handling of the above exception another exception occurred:'
        
        elif reason_type == REASON_TYPE_CAUSE_GROUP:
            title = f'The above {len(exception)} exception was the direct cause of the following exception:'
        
        else:
            title = None
        
        extend.append('\n')
        
        if (title is not None):
            extend = add_trace_title_into(title, highlighter, extend)
            extend.append('\n\n')
        continue
    
    return extend



__all__ = ('ExceptionRepresentationBase',)


class ExceptionRepresentationBase:
    """
    Holds an exception's representation.
    """
    __slots__ = ()
    
    def __new__(cls, exception, frames):
        """
        Creates a new exception representation.
        
        Parameters
        ----------
        exception : `BaseException`
            Exception to represent.
        frames : `None | list<FrameProxyBase>`
            The frames the exception is raised with.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_fields(cls):
        """
        Creates a new exception representation from the given fields.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the exception representation's representation."""
        repr_parts = ['<', type(self).__name__]
        
        self._populate_repr_parts(repr_parts)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _populate_repr_parts(self, repr_parts):
        """
        Helper for ``.__repr__`` to populate its parts with type specific information.
        
        Parameters
        ----------
        repr_parts : `list<str>`
            The representation parts to populate.
        """
        pass
    
    
    def __eq__(self, other):
        """Returns whether the two exception representations are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal(other)
    
    
    def _is_equal(self, other):
        """
        Returns whether the two exception representations are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be same type as `self`.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True



__all__ = ()

from itertools import islice

from ..core import LOOP_TIME
from ..utils import RichAttributeErrorBaseType

from .constants import HOST_INFO_CACHE_TIMEOUT
from .host_info import HostInfo


class HostInfoBasket(RichAttributeErrorBaseType):
    """
    Holds ``HostInfo`` instances and rotates them.
    
    Attributes
    ----------
    expiration : `int`
        Monotonic time determining when the goods in the basket will expire
    
    host_infos : `tuple<HostInfo>`
        A list of the contained host information.
    
    rotation_start_index : `int`
        An index to determine where the next rotation will start. Used to cycle host infos.
    """
    __slots__ = ('expiration', 'host_infos', 'rotation_start_index')
    
    
    def __new__(cls, host_infos):
        """
        Creates a new host info basket.
        
        Parameters
        ----------
        host_infos : `iterable<HostInfo>`
            Host infos to basket up.
        """
        self = object.__new__(cls)
        self.expiration = LOOP_TIME() + HOST_INFO_CACHE_TIMEOUT
        self.host_infos = tuple(host_infos)
        self.rotation_start_index = 0
        return self
    
    
    @classmethod
    def from_address_infos(cls, host_name, address_infos):
        """
        Creates a new host info basket from the given address infos.
        
        Parameters
        ----------
        host_name : `str`
            The host's name.
        
        address_infos : `list<(AddressFamily, SocketKind, int, str, (str, int) | (str, int, int, int))>`
            Address infos returned by `get_address_info`.
        """
        self = object.__new__(cls)
        self.expiration = LOOP_TIME() + HOST_INFO_CACHE_TIMEOUT
        self.host_infos = (*(HostInfo.from_address_info(host_name, address_info) for address_info in address_infos),)
        self.rotation_start_index = 0
        return self
    
    
    def __repr__(self):
        """Returns the host info basket's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # host_infos <- this is the most valuable field
        repr_parts.append(' host_infos = ')
        repr_parts.append(repr(self.host_infos))
        
        # expiration
        repr_parts.append(', expiration = ')
        repr_parts.append(repr(self.expiration))
        
        # rotation_start_index
        repr_parts.append(', rotation_start_index = ')
        repr_parts.append(repr(self.rotation_start_index))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __mod__(self, other):
        """Returns whether the two host infos are alike."""
        if type(self) is not type(other):
            return NotImplemented
        
        return {*self.host_infos} == {*other.host_infos}
    
    
    def __contains__(self, host_info):
        """Returns whether the basket contains the given host info."""
        return host_info in self.host_infos
    
    
    # Note:
    # Do not add `__eq__` and `__hash__` because it is expected that these object change in haste.
    
    
    def is_expired(self):
        """
        Returns whether the host info basket expired.
        
        Returns
        -------
        expired : `bool`
        """
        return self.expiration < LOOP_TIME()
    
    
    def iter_next_rotation(self):
        """
        Gets the next rotated host infos from the basket.
        
        This function is an iterable generator.
        
        Yields
        ------
        result : `list<HostInfo>`
        """
        host_infos = self.host_infos
        length = len(host_infos)
        
        rotation_start_index = self.rotation_start_index
        
        new_rotation_start_index = rotation_start_index + 1
        if new_rotation_start_index == length:
            new_rotation_start_index = 0
        self.rotation_start_index = new_rotation_start_index
        
        yield from islice(host_infos, rotation_start_index, length)
        yield from islice(host_infos, 0, rotation_start_index)



__all__ = ('debug_key',)

import sys, termios, tty
from os import get_blocking, set_blocking
from select import select

from ...tools.keys import KEY_NAMES


def input_once():
    """
    Inputs once from `sys.stdin`.
    
    Returns
    -------
    output : `str`
    """
    input_stream = sys.stdin
    
    input_stream_settings_original = termios.tcgetattr(input_stream)
    input_stream_blocking_original = get_blocking(input_stream.fileno())
    set_blocking(input_stream.fileno(), False)
    tty.setraw(input_stream)
    
    try:
        while True:
            read_file_descriptors, write_file_descriptors, exception_file_descriptors = \
                select([input_stream.fileno()], [], [])
            
            if not read_file_descriptors:
                continue
            
            try:
                content = input_stream.read()
            except (BlockingIOError, InterruptedError):
                continue
            
            if not content:
                continue
            
            return content
    finally:
        termios.tcsetattr(input_stream, termios.TCSADRAIN, input_stream_settings_original)
        set_blocking(input_stream.fileno(), input_stream_blocking_original)


def _escape_character(character):
    """
    Escapes the given character.
    
    Parameters
    ----------
    character : `str`
        The character to escape.
    
    Returns
    -------
    result : `str`
    """
    character_int = ord(character)
    if character_int >= 0x10000 or character_int < 0:
        prefix = '\\U'
        hexadecimal_length = 8
    
    elif character_int >= 0x100:
        prefix = '\\u'
        hexadecimal_length = 4
    
    else:
        prefix = '\\x'
        hexadecimal_length = 2
    
    result = [prefix]
    for index in reversed(range(hexadecimal_length)):
        result.append('0123456789abcdef'[(character_int >> (index << 2)) & 0x0f])
    
    return ''.join(result)


def _escape_string(content):
    """
    Escapes the given string.
    
    Returns
    -------
    escaped : `str`
    """
    return ''.join([_escape_character(character) for character in content])


def _determine_input_name(content):
    """
    Determines the input's content name if its a special character.
    
    Parameters
    ----------
    content : `str`
        Content to get its name of.
    
    Returns
    -------
    name : `str`
    """
    try:
        output = KEY_NAMES[content]
    except KeyError:
        if content.isprintable():
            return content
        
        return _escape_string(content)
    
    return output


def debug_key():
    """
    Prints debug information for a pressed key. Available only on linux.
    """
    if sys.platform != 'linux':
        sys.stdout.write('Linux only.\n')
        return
    
    sys.stdout.write('Please press a key on your keyboard.\n')
    name = _determine_input_name(input_once())
    sys.stdout.write(f'You pressed: {name!s}\n')



__all__ = ()

from warnings import warn

from ..utils import RichAttributeErrorBaseType
from ..web_common.helpers import freeze_headers


class RequestInfo(RichAttributeErrorBaseType):
    """
    Base information representing a request.
    
    Attributes
    ----------
    headers : ``IgnoreCaseMultiValueDictionary``
        The respective request's headers.
    
    method : `str`
        The respective request's method.
    
    original_url : ``URL``
        The url given to request.
    
    url : ``URL``
        The requested url without fragments. Can be same as `original_url`.
    """
    __slots__ = ('headers', 'method', 'original_url', 'url',)
    
    def __new__(cls, headers, method, original_url, url):
        """
        Creates a new request info representing the given request.
        
        Parameters
        ----------
        headers : ``IgnoreCaseMultiValueDictionary``
            The respective request's headers.
        
        method : `str`
            The respective request's method.
        
        original_url : ``URL``
            The url given to request.
        
        url : ``URL``
            The requested url without fragments. Can be same as `original_url`.
        """
        self = object.__new__(cls)
        self.headers = headers
        self.method = method
        self.original_url = original_url
        self.url = url
        return self
    
    
    def __repr__(self):
        """Returns the request info's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # method
        repr_parts.append(', method = ')
        repr_parts.append(repr(self.method))
        
        # url
        url = self.url
        repr_parts.append(', url = ')
        repr_parts.append(repr(url))
        
        # original_url
        original_url = self.original_url
        if (url != original_url):
            repr_parts.append(', original_url = ')
            repr_parts.append(repr(original_url))
        
        # headers
        headers = self.headers
        if headers:
            repr_parts.append(', headers = ')
            repr_parts.append(repr(headers))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two connection infos are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # headers
        if self.headers != other.headers:
            return False
        
        # method
        if self.method != other.method:
            return False
        
        # original_url
        if self.original_url != other.original_url:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the request info's hash value."""
        hash_value = 0
        
        # headers
        headers_frozen = freeze_headers(self.headers)
        if (headers_frozen is not None):
            hash_value ^= hash(headers_frozen)
        
        # method
        hash_value ^= hash(self.method)
        
        # original_url
        original_url = self.original_url
        hash_value ^= hash(original_url)
        
        # url
        url = self.url
        if (url != original_url):
            hash_value ^= hash(url)
        
        return hash_value
    
    
    @property
    def real_url(self):
        """
        Deprecated and will be removed in 2025 August.
        """
        warn(
            (
                f'The `proxy_url` and `proxy_headers` parameters in `{type(self).__name__}.__new__` are moved to be '
                f'keyword only. Support for positional is deprecated and will be removed in 2025 August.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.original_url



__all__ = (
    'add_highlighted_part_into', 'add_highlighted_parts_into', 'get_token_type_and_repr_mode_for_variable',
    'iter_highlight_code_lines',
)

from .constants import BUILTIN_CONSTANTS, BUILTIN_EXCEPTIONS, BUILTIN_VARIABLES
from .parser_context import HighlightParserContext
from .token import Token
from .token_types import (
    TOKEN_TYPE_IDENTIFIER_BUILTIN_CONSTANT, TOKEN_TYPE_IDENTIFIER_BUILTIN_EXCEPTION,
    TOKEN_TYPE_IDENTIFIER_BUILTIN_VARIABLE, TOKEN_TYPE_IDENTIFIER_VARIABLE, TOKEN_TYPE_NON_SPACE_UNIDENTIFIED,
    TOKEN_TYPE_NUMERIC_FLOAT, TOKEN_TYPE_NUMERIC_INTEGER, TOKEN_TYPE_STRING_BINARY, TOKEN_TYPE_STRING_UNICODE
)


def iter_highlight_code_lines(lines, formatter_context):
    """
    Matches the given python code lines and iterates it's formatted representation.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines to format.
    formatter_context : ``HighlightFormatterContext``
        Context to use for highlighting.
    
    Yields
    ------
    content : `str`
    """
    context = HighlightParserContext(lines)
    context.match()
    yield from context.generate_highlighted(formatter_context)


def add_highlighted_part_into(token_type, part, highlighter, into):
    """
    Adds a highlighted part extending the given list of strings.
    
    Parameters
    ----------
    token_type : `int`
        Token type identifier.
    
    part : `str`
        The part to add.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Returns
    -------
    into : `list<str>`
    """
    if (highlighter is None):
        into.append(part)
    else:
        into.extend(highlighter.generate_highlighted(Token(token_type, part)))
    
    return into
    

def add_highlighted_parts_into(producer, highlighter, into):
    """
    Iterates over a producer and applies highlighter for each produced item extending the given list of strings with
    them.
    
    Parameters
    ----------
    producer : `iterable<(int, str)>`
        Iterable to produce token_type - part items.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Yields
    ------
    into : `list<str>`
    """
    for token_type, part in producer:
        into = add_highlighted_part_into(token_type, part, highlighter, into)
    
    return into


def get_token_type_and_repr_mode_for_variable(variable):
    """
    Gets the token type for the given variable.
    
    Parameters
    ----------
    variable : `object`
        The variable to get token type for.
    
    Returns
    -------
    token_type, use_name : `(int, bool)`
        What token-type should be used for highlighting the given variable & whether the name of the variable should
        be used instead of its representation to show it.
    """
    while True:
        if isinstance(variable, str):
            if (type(variable).__repr__ is str.__repr__):
                token_type = TOKEN_TYPE_STRING_UNICODE
                use_name = False
                break
        
        elif isinstance(variable, bytes):
            if (type(variable).__repr__ is bytes.__repr__):
                token_type = TOKEN_TYPE_STRING_BINARY
                use_name = False
                break
        
        elif isinstance(variable, int) and (not isinstance(variable, bool)):
            if (type(variable).__repr__ is int.__repr__):
                token_type = TOKEN_TYPE_NUMERIC_INTEGER
                use_name = False
                break
        
        elif isinstance(variable, float):
            if (type(variable).__repr__ is float.__repr__):
                token_type = TOKEN_TYPE_NUMERIC_FLOAT
                use_name = False
                break
        
        else:
            try:
                hash(variable)
            except TypeError:
                pass
            
            else:
                if variable in BUILTIN_CONSTANTS:
                    token_type = TOKEN_TYPE_IDENTIFIER_BUILTIN_CONSTANT
                    use_name = False
                    break
                
                elif variable in BUILTIN_VARIABLES:
                    token_type = TOKEN_TYPE_IDENTIFIER_BUILTIN_VARIABLE
                    use_name = True
                    break
                
                elif variable in BUILTIN_EXCEPTIONS:
                    token_type = TOKEN_TYPE_IDENTIFIER_BUILTIN_EXCEPTION
                    use_name = True
                    break
        
        if isinstance(variable, type):
            token_type = TOKEN_TYPE_IDENTIFIER_VARIABLE
            use_name = True
            break
        
        token_type = TOKEN_TYPE_NON_SPACE_UNIDENTIFIED
        use_name = False
        break
    
    return token_type, use_name



__all__ = ()

from ...utils import RichAttributeErrorBaseType

from ..quoting import quote, unquote

from .constants import URL_PART_FLAG_DECODED_SET, URL_PART_FLAG_ENCODED_SET


class URLPartBase(RichAttributeErrorBaseType):
    """
    Represents an url's part.
    
    Attributes
    ----------
    _decoded : `None | str`
        Decoded value.
    
    _encoded : `None | str`
        Encoded value.
    
    _flags : `int`
        Bitwise flags representing how the object is constructed.
    """
    __slots__ = ('_decoded', '_encoded', '_flags')
    
    
    def __new__(cls):
        """
        Use either ``.create_from_decoded`` or ``.create_from_encoded`` instead.
        
        Raises
        ------
        RuntimeError
        """
        raise RuntimeError(
            'Use either `.create_from_decoded` or `.create_from_encoded` instead.'
        )
    
    
    def __repr__(self):
        """Returns the url path's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # decoded
        decoded = self.decoded
        if (decoded is not None):
            repr_parts.append(' decoded = ')
            repr_parts.append(repr(decoded))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two url parts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # decoded
        if self.decoded != other.decoded:
            return False
        
        return True
    
    
    @staticmethod
    def _encode(value):
        """
        Encodes the given value.
        
        Parameters
        ----------
        value : `None | str`
            The value to encode.
        
        Returns
        -------
        value : `None | str`
        """
        return quote(value)
    
    
    @staticmethod
    def _decode(value):
        """
        Decodes the given value.
        
        Parameters
        ----------
        value : `None | str`
            The value decode.
        
        Returns
        -------
        value : `None | str`
        """
        return unquote(value)
    
    
    @classmethod
    def create_from_decoded(cls, decoded):
        """
        Creates a new url part from a decoded value.
        
        Parameters
        ----------
        decoded : `None | str`
            Already decoded value.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._decoded = decoded
        self._encoded = None
        self._flags = URL_PART_FLAG_DECODED_SET
        return self
    
    
    @classmethod
    def create_from_encoded(cls, encoded):
        """
        Creates a new url part from a encoded value.
        
        Parameters
        ----------
        encoded : `None | str`
            Encoded value.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._decoded = None
        self._encoded = encoded
        self._flags = URL_PART_FLAG_ENCODED_SET
        return self


    @property
    def decoded(self):
        """
        Returns the part's decoded string representation.
        
        Returns
        -------
        decoded : `None | str`
            Decoded value.
        
        Returns
        -------
        self : `instance<cls>`
        """
        flags = self._flags
        if flags & URL_PART_FLAG_DECODED_SET:
            decoded = self._decoded
        
        else:
            if flags & URL_PART_FLAG_ENCODED_SET:
                decoded = self._decode(self._encoded)
            else:
                decoded = None
            
            self._decoded = decoded
            self._flags = flags | URL_PART_FLAG_DECODED_SET
        
        return decoded
    
    
    @property
    def encoded(self):
        """
        Returns the part's decoded string representation.
        
        Returns
        -------
        decoded : `None | str`
            Decoded value.
        """
        flags = self._flags
        if flags & URL_PART_FLAG_ENCODED_SET:
            encoded = self._encoded
        
        else:
            encoded = self._encode(self.decoded)
            
            self._encoded = encoded
            self._flags = flags | URL_PART_FLAG_ENCODED_SET
        
        return encoded



__all__ = ('ignore_frame',)

import os
from warnings import warn

from ..analyzer import CallableAnalyzer
from ..docs import copy_docs

from .formatters import format_builtin


DEFAULT_IGNORED_FRAMES = os.environ.get('SCARLETIO_TRACE_DEFAULT_IGNORED_FRAME', '').casefold() not in ('0', 'false')

IGNORED_FRAME_LINES = {}


def ignore_frame(file_name, name, line):
    """
    When rendering an exception traceback, specified frames can be added to being stopped from rendering.
    
    Parameters
    ----------
    file_name : `str`
        The name of the respective file.
    name : `str`
        The name of the respective function.
    line : `str`
        The respective line's stripped content.
    """
    key = (file_name, name)
    
    try:
        lines = IGNORED_FRAME_LINES[key]
    except KeyError:
        lines = set()
        IGNORED_FRAME_LINES[key] = lines
    
    lines.add(line)


def should_keep_frame_from_defaults(frame_proxy):
    """
    Returns whether the given frame should be kept and rendered.
    
    Called before before calling it's `should_keep_frame_from_filter`.
    
    Parameters
    ----------
    frame_proxy : ``FrameProxyBase``
        Frame proxy to check whether a frame should be ignored.
    
    Returns
    -------
    should_keep_frame : `bool`
        Whether the frame should be ignored.
    """
    try:
        lines = IGNORED_FRAME_LINES[frame_proxy.file_name, frame_proxy.name]
    except KeyError:
        return True
    
    if frame_proxy.line not in lines:
        return True
    
    return False


if not DEFAULT_IGNORED_FRAMES:
    @copy_docs(should_keep_frame_from_defaults)
    def should_keep_frame_from_defaults(frame_proxy):
        return True


def should_keep_frame_from_filter(frame_proxy, filter):
    """
    Returns whether a frame should be kept based on the given filter.
    
    Parameters
    ----------
    frame_proxy : ``FrameProxyBase``
        Frame proxy to check whether a frame should be ignored.
    filter : `FunctionType`
        Filter to check the frame.
    
    Returns
    -------
    should_keep_frame : `bool`
        Whether the frame should be ignored.
    """
    if CallableAnalyzer(filter).get_non_reserved_positional_parameter_count() == 1:
        return filter(frame_proxy)
    
    warn(
        (
            f'`should_keep_frame` now passes `1` parameter to `filter`, but a filter given with different '
            f'amount. Please update your filter function to accept `1`. Got {format_builtin(filter)!s}. '
            f'4 parameter version will be removed in 2025 January.'
        ),
        FutureWarning,
    )
    return filter(frame_proxy.file_name, frame_proxy.name, frame_proxy.line_index, frame_proxy.line)


def should_keep_frame(frame_proxy, *, filter = None):
    """
    Returns whether the given frame should be kept and rendered.
    
    Parameters
    ----------
    frame_proxy : ``FrameProxyBase``
        Frame proxy to check whether a frame should be ignored.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
    
    Returns
    -------
    should_keep_frame : `bool`
    """
    if not should_keep_frame_from_defaults(frame_proxy):
        return False
    
    if (filter is not None) and (not should_keep_frame_from_filter(frame_proxy, filter)):
        return False
    
    return True


# ---- Ignore frames / module_rich_attribute_error ----

from .. import rich_attribute_error as module_rich_attribute_error

ignore_frame(
    module_rich_attribute_error.__spec__.origin, '__getattr__', 'raise AttributeError(self, attribute_name)'
)


del module_rich_attribute_error


# ---- Ignore frames / function_tools ----

from .. import function_tools as module_function_tools

ignore_frame(
    module_function_tools.__spec__.origin,
    '__call__',
    'return self.function(*self.positional_parameters)',
)
ignore_frame(
    module_function_tools.__spec__.origin,
    '__call__',
    'return self.function(*self.positional_parameters, **keyword_parameters)',
)

del module_function_tools



from .cache_constants import *
from .console_input import *
from .expression_info import *
from .expression_key import *
from .file_info import *
from .line_cache_session import *
from .line_info import *
from .parsing import *


# We do not want to pass them down, except that one.
__all__ = (
    'add_console_input',
)



__all__ = ()

from ..utils import RichAttributeErrorBaseType


class ProtocolBasket(RichAttributeErrorBaseType):
    """
    Basket to hold protocols with details about them.
    
    Attributes
    ----------
    available : `None | list<(AbstractProtocolBase, float, int)>`
        Available protocols.
    
    connection_key : ``ConnectionKey``
        Key representing the connection.
    
    used : `None | set<AbstractProtocolBase>`
        The protocols in use.
    """
    __slots__ = ('available', 'connection_key', 'used')
    
    
    def __new__(cls, connection_key):
        """
        Creates a new protocol basket.
        
        Parameters
        ----------
        connection_key : ``ConnectionKey``
            Key representing the connection.
        """
        self = object.__new__(cls)
        self.available = None
        self.connection_key = connection_key
        self.used = None
        return self
    
    
    def __repr__(self):
        """Returns the protocol basket's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # connection_key
        repr_parts.append(' connection_key = ')
        repr_parts.append(repr(self.connection_key))
        
        # available
        available = self.available
        if (available is not None):
            repr_parts.append(', available = ')
            repr_parts.append(repr(available))
        
        # used
        used = self.used
        if (used is not None):
            repr_parts.append(' used = ')
            repr_parts.append(repr(used))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two protocol baskets are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # available
        if self.available != other.available:
            return False
        
        # connection_key
        if self.connection_key != other.connection_key:
            return False
        
        # used
        if self.used != other.used:
            return False
        
        return True
    
    
    def __bool__(self):
        """Returns whether the protocol basket stores any protocols."""
        return (self.available is not None) or (self.used is not None)
    
    
    def clean_up_expired_protocols(self, now):
        """
        Cleans up the protocols that are expired.
        
        Parameters
        ----------
        now : `bool`
            The current monotonic time.
        """
        available = self.available
        if available is None:
            return
        
        for index in reversed(range(len(available))):
            protocol, expiration, performed_requests = available[index]
            if expiration > now:
                continue
            
            del available[index]
            
            transport = protocol.get_transport()
            if self.connection_key.secure and (transport is not None):
                transport.abort()
        
        if not available:
            self.available = None
    
    
    def close(self):
        """
        Closes all the protocols in the basket.
        """
        available = self.available
        if (available is not None):
            self.available = None
            for protocol, expiration, performed_requests in available:
                protocol.close()
            
        used = self.used
        if (used is not None):
            self.used = None
            for protocol in used:
                protocol.close()
    
    
    def pop_available_protocol(self, now):
        """
        Pops an available protocol.
        
        Parameters
        ----------
        now : `float`
            The current monotonic time.
        
        Returns
        -------
        protocol : `None | AbstractProtocolBase`
            Available protocol.
        
        performed_requests : `int`
            The amount of performed requests on the protocol.
        """
        available = self.available
        if available is None:
            return None, 0
        
        while available:
            # get the item with lowest expiration
            iterator = iter(available)
            hit = next(iterator)
            hit_index = 0
            
            for index, item in enumerate(iterator, 1):
                if item[1] < hit[1]:
                    hit = item
                    hit_index = index
            
            del available[hit_index]
            protocol, expiration, performed_requests = hit
            if (protocol.get_transport() is None):
                continue
            
            if now > expiration:
                transport = protocol.get_transport()
                protocol.close()
                if self.connection_key.secure and (transport is not None):
                    transport.abort()
                continue
            
            break
        
        else:
            protocol = None
            performed_requests = 0
        
        if not available:
            self.available = None
        
        return protocol, performed_requests
    
    
    def add_used_protocol(self, protocol):
        """
        Adds a used protocol to the basket.
        
        Parameters
        ----------
        protocol : ``AbstractProtocolBase``
            The protocol to remove.
        """
        used = self.used
        if used is None:
            used = set()
            self.used = used
        
        used.add(protocol)
    
    
    def remove_used_protocol(self, protocol):
        """
        Removes a used protocol.
        
        Parameters
        ----------
        protocol : ``AbstractProtocolBase``
            The protocol to remove.
        """
        used = self.used
        if used is None:
            return
        
        try:
            used.discard(protocol)
        except KeyError:
            pass
        
        if not used:
            self.used = None
    
    
    def add_available_protocol(self, protocol, now, keep_alive_timeout, performed_requests):
        """
        Adds an available protocol to the basket.
        
        Parameters
        ----------
        protocol : ``AbstractProtocolBase`
            The protocol to add.
        
        now : `float`
            The current monotonic time.
        
        keep_alive_timeout : `float`
            How long the connection can be reused.
        
        performed_requests : `int`
            The amount of performed requests on the connection.
        """
        available = self.available
        if available is None:
            available = []
            self.available = available
        
        available.append((protocol, now + keep_alive_timeout, performed_requests))
    
    
    def get_closest_expiration(self):
        """
        Returns the closest expiration of the protocol basket.
        
        Returns
        -------
        closest_expiration : `float`
            Returns `-1` on failure.
        """
        available = self.available
        if available is None:
            return -1.0
        
        return min(item[1] for item in available)



__all__ = ()

from itertools import islice


# Pulled from C 3.11
#
# args = Py_BuildValue("(O(OiiNii))", errmsg, tok->filename, tok->lineno,
#                      col_offset, errtext, tok->lineno, end_col_offset);
# if (args) {
#     PyErr_SetObject(PyExc_SyntaxError, args);
#
# file_name is annotated as `O`, so it probably can be `None` as well.
# also: line_number == end_line_number, so we can ignore it.
#
# On older Python versions `line` can also be `None`. How stupid!
#
#
# At some cases exception is built differently tho:
#
# PyErr_Format(PyExc_SyntaxError, DUPLICATE_ARGUMENT, name);
# SET_ERROR_LOCATION(st->st_filename, loc);
#
# Here the exception's details are omitted instead they are set as slots only.
# Because we do not use the slots instead pull the details from its parameters, we have to fix exception up.

def _are_details_valid(details):
    """
    Checks whether the syntax error details are valid.
    
    Parameters
    ----------
    details : `(object, object, object, object) | (object, object, object, object, object, object)`
        A tuple of length 4 or 6.
    
    Returns
    -------
    details_valid : `bool`
    """
    file_name = details[0]
    line_number = details[1]
    offset = details[2]
    line = details[3]
    
    if (file_name is not None) and (not isinstance(file_name, str)):
        return False
    
    if not isinstance(line_number, int):
        return False
    
    if not isinstance(offset, int):
        return False
    
    if (line is not None) and (not isinstance(line, str)):
        return False
    
    if len(details) == 4:
        return True
    
    end_line_number = details[4]
    end_offset = details[5]
    
    if not isinstance(end_line_number, int):
        return False
    
    if not isinstance(end_offset, int):
        return False
    
    return True
    

def is_syntax_error(exception):
    """
    Returns whether the given exception is a syntax error with the expected structure.
    
    Parameters
    ----------
    exception : ``BaseException``
        The exception to check.
    
    Returns
    -------
    is_syntax_error : `bool`
    """
    if not issubclass(type(exception), SyntaxError):
        return False
    
    exception_parameters = type(exception).args.__get__(exception, type(exception))
    if (not isinstance(exception_parameters, tuple)) or (len(exception_parameters) != 2):
        return False
    
    message, details = exception_parameters
    if (message is not None) and (not isinstance(message, str)):
        return False
    
    if (not isinstance(details, tuple)) or (len(details) not in (4, 6)):
        return False
    
    return _are_details_valid(details)


def _get_details_from_slots(syntax_error):
    """
    Gets the syntax error's details from its members instead of its parameters.
    
    Parameters
    ----------
    syntax_error : ``SyntaxError``
        Syntax error to get details of.
    
    Returns
    -------
    details : `(object, object, object, object) | (object, object, object, object, object, object)`
        A tuple of length 4 or 6.
    """
    details = (syntax_error.filename, syntax_error.lineno, syntax_error.offset, syntax_error.text)
    
    if hasattr(SyntaxError, 'end_lineno') and hasattr(SyntaxError, 'end_offset'):
        details = (*details, syntax_error.end_lineno, syntax_error.end_offset)
    
    return details


def is_syntax_error_details_omitted(exception):
    """
    Returns whether the syntax error's has ots details omitted, but they are set in slots instead.
    
    Parameters
    ----------
    exception : ``SyntaxError``
        The exception to check.
    
    Returns
    -------
    details_omitted : `bool`
    """
    exception_parameters = type(exception).args.__get__(exception, type(exception))
    if (not isinstance(exception_parameters, tuple)) or (len(exception_parameters) != 1):
        return False
    
    message, = exception_parameters
    if (message is not None) and (not isinstance(message, str)):
        return False
    
    details = _get_details_from_slots(exception)
    return _are_details_valid(details)


def fixup_syntax_error_details_omitted(syntax_error):
    """
    Fixes up the syntax error's details if they are omitted.
    
    Parameters
    ----------
    syntax_error : ``SyntaxError``
        Syntax error to check.
    """
    syntax_error.args = (syntax_error.args[0], _get_details_from_slots(syntax_error))


def fixup_syntax_error_line_from_buffer(syntax_error, buffer):
    """
    Tries to fix up the syntax error's missing line.
    
    Should be only called if ``is_syntax_error`` returned `True` on the given exception.
    
    Parameters
    ----------
    syntax_error : ``SyntaxError``
        Respective SyntaxError
    buffer : `list` of `str`
        Buffer containing the respective lines of the file.
    """
    message, details = syntax_error.args
    line = details[3]
    if line is not None:
        return
    
    line_number = details[1]
    buffer_length = len(buffer)
    # `line_number` means `line_index + 1`
    if (line_number > buffer_length) or (line_number < 1):
        return
    
    line = buffer[line_number - 1]
    syntax_error.args = (message, (*islice(details, 0, 3), line, *islice(details, 4, 6)))


def right_strip_syntax_error_line(syntax_error):
    """
    Right strips the syntax error's line. This can be useful when comparing two syntax errors, but one has new line
    character at the end.
    
    Should be only called if ``is_syntax_error`` returned `True` on the given exception.
    
    Parameters
    ----------
    syntax_error : `SyntaxError`
        The syntax error to strip.
    """
    message, (file_name, line_number, offset, line, *end) = syntax_error.args
    if line is None:
        return
    
    line = line.rstrip()
    
    syntax_error.args = (message, (file_name, line_number, offset, line, *end))



__all__ = ('SSLFingerprint',)

from hashlib import md5, sha1

from ..utils import RichAttributeErrorBaseType

from .constants import HASH_FUNCTION_BY_DIGEST_LENGTH


class SSLFingerprint(RichAttributeErrorBaseType):
    """
    HTTP fingerprinting can be used to automate information systems and security audits.
    Automated security testing tools can use HTTP fingerprinting to narrow down the set of tests required,
    based on the specific platform or the specific web server being audited.
    
    Attributes
    ----------
    hash_function : `function`
        Hash function used by the fingerprint.
    value : `bytes`
        The fingerprint's value.
    """
    __slots__ = ('hash_function', 'value')
    
    def __new__(cls, value):
        """
        Creates a new fingerprint with the given parameters.
        
        Parameters
        ----------
        value : `bytes`
            SSLFingerprint value.
        
        Raises
        ------
        ValueError
            - If `fingerprint`'s length is not any of the expected ones.
            - If the detected `hash_function` is `md5` / `sha1`.
        """
        value_length = len(value)
        
        try:
            hash_function = HASH_FUNCTION_BY_DIGEST_LENGTH[value_length]
        except KeyError:
            raise ValueError(
                f'`fingerprint` has invalid length, got: value_length = {value_length!r}; value = {value!r}.'
            ) from None
        
        if hash_function is md5 or hash_function is sha1:
            raise ValueError(
                f'`md5` and `sha1` are insecure, use `sha256`. '
                f'Got value_length = {value_length!r}; value = {value!r}; hash_function = {hash_function.__name__}.'
            )
        
        self = object.__new__(cls)
        self.hash_function = hash_function
        self.value = value
        return self
    
    
    def check(self, protocol):
        """
        Checks whether the given protocol's ssl data matches the fingerprint.
        
        Parameters
        ----------
        protocol : ``AbstractProtocolBase``
            Asynchronous protocol implementation.
        
        Raises
        ------
        ValueError
            If the fingerprint don't match.
        """
        ssl_object = protocol.get_extra_info('ssl_object')
        if ssl_object is None:
            return
        
        certificate = ssl_object.getpeercert(True)
        received_value = self.hash_function(certificate).digest()
        own_value = self.value
        if received_value == own_value:
            return
        
        address = protocol.get_extra_info('peer_name')
        if address is None:
            host = 'unknown'
            port = 'unknown'
        else:
            host = address[0]
            port = str(address[1])
        
        raise ValueError(
            f'The expected fingerprint: {own_value!r} not matches the received; received = {received_value!r}; '
            f'host = {host}; port = {port}.'
        )
    
    
    def __eq__(self, other):
        """Returns whether the two fingerprints are equal.""" 
        if type(self) is not type(other):
            return NotImplemented
        
        return self.value == other.value
    
    
    def __repr__(self):
        """Returns the fingerprint's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' value = ')
        repr_parts.append(repr(self.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


    def __hash__(self):
        """Returns the fingerprint's hash value."""
        return hash(self.value)



__all__ = ()

from ..highlight import (
    HIGHLIGHT_TOKEN_TYPES, add_highlighted_part_into, add_highlighted_parts_into, iter_highlight_code_lines
)
from ..highlight.constants import (
    BUILTIN_EXCEPTION_NAMES, BUILTIN_VARIABLE_NAMES, MAGIC_FUNCTION_NAMES, MAGIC_VARIABLE_NAMES
)

from .exception_representation import (
    ExceptionRepresentationAttributeError, ExceptionRepresentationGeneric, ExceptionRepresentationSyntaxError
)


def render_exception_proxy_into(exception, highlighter, into):
    """
    Renders the given frame groups.
    
    Parameters
    ----------
    exception : ``ExceptionProxyBase``
        The frame groups to render.
    
    highlighter : `None`, ``HighlightFormatterContext``
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to render the representation into.
    
    Returns
    -------
    into : `list<str>`
    """
    into = render_frame_groups_into(exception.frame_groups, highlighter, into)
    into = render_exception_representation_into(exception.exception_representation, highlighter, into)
    return into


def render_frame_groups_into(frame_groups, highlighter, into):
    """
    Renders the given frame groups.
    
    Parameters
    ----------
    frame_groups : `None | list<FrameGroup>`
        The frame groups to render.
    
    highlighter : `None`, ``HighlightFormatterContext``
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to render the representation into.
    
    Returns
    -------
    into : `list<str>`
    """
    if (frame_groups is not None):
        for frame_group in frame_groups:
            into = render_frame_group_into(frame_group, highlighter, into)
    
    return into


def _build_frames_repeated_line(frame_count, repeat_count):
    """
    Builds the `frames repeated` line.
    
    Parameters
    ----------
    frame_count : `lint`
        The amount of repeated frames.
    repeat_count : `int`
        How much times the frames were repeated.
    
    Returns
    -------
    line : `str`
    """
    parts = []
    parts.append('[Following ')
    parts.append(str(frame_count))
    parts.append(' frame')
    parts.append('s were' if frame_count > 1 else ' was')
    parts.append(' repeated ')
    parts.append(str(repeat_count))
    parts.append(' times]')
    
    return ''.join(parts)


def render_frame_group_into(frame_group, highlighter, into):
    """
    Renders the frame group.
    
    Parameters
    ----------
    frame_group : ``FrameGroup``
        The frame group to render.
    
    highlighter : `None`, ``HighlightFormatterContext``
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to render the representation into.
    
    Returns
    -------
    into : `list<str>`
    """
    frames = frame_group.frames
    if frames is None:
        # Empty frame group, should not happen.
        return into
    
    repeat_count = frame_group.repeat_count
    if (repeat_count > 1):
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_FRAME_REPEAT,
            _build_frames_repeated_line(len(frames), repeat_count),
            highlighter,
            into,
        )
        into.append('\n')
    
    for frame in frames:
        into = render_frame_proxy_into(frame, highlighter, into)
    
    if (repeat_count > 1):
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_FRAME_REPEAT,
            '[End of repeated frames]',
            highlighter,
            into,
        )
        into.append('\n')
    
    return into


def render_frame_proxy_into(frame, highlighter, into):
    """
    Produces each part of the frame to render.
    
    This method is an iterable generator.
    
    Parameters
    ----------
    frame : ``FrameProxyBase``
        The frame to render.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Yields
    ------
    into : `list<str>`
    """
    lines = frame.lines
    
    add_highlighted_parts_into(
        _produce_file_location(frame.file_name, frame.line_index, frame.name, len(lines)),
        highlighter,
        into,
    )
    
    if lines:
        if (highlighter is None):
            for line in lines:
                into.append('    ')
                into.append(line)
                into.append('\n')
        
        else:
            into.extend(iter_highlight_code_lines([f'    {line}\n' for line in lines], highlighter))
    
    return into


def _produce_file_location(file_name, line_index, name, line_count):
    """
    Produces file location part and their tokens.
    
    Parameters
    ----------
    file_name : `str`
        Path of the respective file.
    line_index : int`
        The respective line's index.
    name : `str`
        The respective functions name.
    line_count : `int`
        How much lines is the expression at the location.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION, '  File '
    
    if file_name:
        file_name = file_name.replace('"', '\\"') # Add escapes into the file name
        yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_PATH, f'"{file_name!s}"'
    else:
        yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION, 'unknown location'
    
    # if its multiple lines add `around` word
    if line_count > 1:
        yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION, ', around line '
    else:
        yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION, ', line '
    yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_LINE_NUMBER, str(line_index + 1)
    
    if name:
        yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION, ', in '
        yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_NAME, name
    
    yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_LINEBREAK, '\n'


def add_trace_title_into(title, highlighter, into):
    """
    Adds trace title into the given list of strings.
    
    Parameters
    ----------
    title : `str`
        The title to add.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Returns
    -------
    into : `list<str>`
    """
    return add_highlighted_part_into(HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE, title, highlighter, into)


def _render_exception_representation_generic_into(exception_representation, highlighter, into):
    """
    Renders a generic exception's representation.
    
    Parameters
    ----------
    exception_representation : ``ExceptionRepresentationGeneric``
        The exception representation to render.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Returns
    -------
    into : `list<str>`
    """
    into = add_highlighted_part_into(
        HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
        exception_representation.representation,
        highlighter,
        into,
    )
    into.append('\n')
    return into


def _render_exception_representation_syntax_error_into(exception_representation, highlighter, into):
    """
    Renders a syntax error exception's representation.
    
    Parameters
    ----------
    exception_representation : ``ExceptionRepresentationSyntaxError``
        The exception representation to render.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Returns
    -------
    into : `list<str>`
    """
    # location
    into = add_highlighted_parts_into(
        _produce_file_location(exception_representation.file_name, exception_representation.line_index, '', 1),
        highlighter,
        into,
    )
    
    line = exception_representation.line
    if line:
        # Add line
        into.append('    ')
        if highlighter is None:
            into.append(line)
        else:
            into.extend(iter_highlight_code_lines([line], highlighter))
        into.append('\n')
        
        # Add pointer
        pointer_length = exception_representation.pointer_length
        if pointer_length:
            into.append(' ' * (4 + exception_representation.pointer_start_offset))
            into = add_highlighted_part_into(
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR, '^' * pointer_length, highlighter, into
            )
            into.append('\n')
    
    # type & message
    representation = exception_representation.type_name
    message = exception_representation.message
    if message:
        representation = f'{representation!s}: {message!s}'
    
    into = add_highlighted_part_into(
        HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR, representation, highlighter, into
    )
    into.append('\n')
    return into


def _produce_variable_name_only(variable_name):
    """
    Produces a variable name such as: `variable_name`.
    Also it them withing grace characters.
    
    Parameters
    ----------
    variable_name : `str`
        Variable name to produce.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    if variable_name in BUILTIN_VARIABLE_NAMES:
        token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_BUILTIN_VARIABLE
    elif variable_name in BUILTIN_EXCEPTION_NAMES:
        token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_BUILTIN_EXCEPTION
    else:
        token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_VARIABLE
    
    yield token_type, variable_name


def _produce_attribute_name_only(attribute_name):
    """
    Produces a variable name such as: `.attribute_name`.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute name to produce.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_SPECIAL_OPERATOR_ATTRIBUTE, '.'
    
    if attribute_name in MAGIC_FUNCTION_NAMES:
        token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_MAGIC_FUNCTION
    elif attribute_name in MAGIC_VARIABLE_NAMES:
        token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_MAGIC_VARIABLE
    else:
        token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_ATTRIBUTE
    
    yield token_type, attribute_name


def _produce_variable_attribute_access_only(variable_name, attribute_name):
    """
    Produces a variable attribute access such as: `variable_name.attribute_name`.
    
    Parameters
    ----------
    variable_name : `str`
        Variable name to produce.
    attribute_name : `str`
        The attribute name to produce.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    yield from _produce_variable_name_only(variable_name)
    yield from _produce_attribute_name_only(attribute_name)


def _produce_grave_wrapped(producer):
    """
    Wraps the given `iterable` within grave characters.
    
    Parameters
    ----------
    producer : `iterable<(int, str)>`
        The producer to wrap between grave characters.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_EXCEPTION_REPR_GRAVE_AFFIX, '`'
    yield from producer
    yield HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_EXCEPTION_REPR_GRAVE_AFFIX, '`'


def _produce_variable_name(variable_name):
    """
    Produces a variable name such as: `variable_name`.
    Also it them withing grace characters.
    
    Parameters
    ----------
    variable_name : `str`
        Variable name to produce.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    yield from _produce_grave_wrapped(_produce_variable_name_only(variable_name))


def _produce_attribute_name(attribute_name):
    """
    Produces a variable name such as: `.attribute_name`.
    Also wraps it withing grace characters.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute name to produce.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    yield from _produce_grave_wrapped(_produce_attribute_name_only(attribute_name))


def _produce_variable_attribute_access(variable_name, attribute_name):
    """
    Produces a variable attribute access such as: `variable_name.attribute_name`.
    Also wraps them withing grace characters.
    
    Parameters
    ----------
    variable_name : `str`
        Variable name to produce.
    attribute_name : `str`
        The attribute name to produce.
    
    Yields
    ------
    token_type : `int`
        The part's type.
    part : `str`
        Part to render
    """
    yield from _produce_grave_wrapped(_produce_variable_attribute_access_only(variable_name, attribute_name))


def _render_exception_representation_attribute_error_into(exception_representation, highlighter, into):
    """
    Renders a syntax error exception's representation.
    
    Parameters
    ----------
    exception_representation : ``ExceptionRepresentationAttributeError``
        The exception representation to render.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Returns
    -------
    into : `list<str>`
    """
    attribute_name = exception_representation.attribute_name
    attribute_unset = exception_representation.suggestion_attribute_unset
    familiar_attribute_names = exception_representation.suggestion_familiar_attribute_names
    matching_variable_exists = exception_representation.suggestion_matching_variable_exists
    variable_names_with_attribute = exception_representation.suggestion_variable_names_with_attribute
    lines_rendered = 0
    
    into = add_highlighted_part_into(
        HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
        exception_representation.type_name,
        highlighter,
        into,
    )
    into = add_highlighted_part_into(
        HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
        ': ',
        highlighter,
        into,
    )
    
    into = add_highlighted_parts_into(
        _produce_grave_wrapped([
            (
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_EXCEPTION_REPR_GRAVE_FILLING,
                exception_representation.instance_type_name,
            ),
        ]),
        highlighter,
        into,
    )
    
    into = add_highlighted_part_into(
        HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
        (' does not have its attribute ' if attribute_unset else ' has no attribute '),
        highlighter,
        into,
    )
    into = add_highlighted_parts_into(
        _produce_attribute_name(attribute_name),
        highlighter,
        into,
    )
    into = add_highlighted_part_into(
        HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
        ' set.' if attribute_unset else '.',
        highlighter,
        into,
    )
    
    into.append('\n')
    lines_rendered += 1
    
    if attribute_unset and matching_variable_exists:
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            'Perhaps you meant to use the ',
            highlighter,
            into,
        )
        
        into = add_highlighted_parts_into(
            _produce_variable_name(attribute_name),
            highlighter,
            into,
        )
    
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            ' variable instead?',
            highlighter,
            into,
        )
    
        into.append('\n')
        lines_rendered += 1
    
    if attribute_unset and (not matching_variable_exists):
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            'Please review its constructors whether they are omitting setting it.',
            highlighter,
            into,
        )
        
        into.append('\n')
        lines_rendered += 1
    
    if (not attribute_unset) and matching_variable_exists:
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            'Did you mean to use the ',
            highlighter,
            into,
        )
        
        into = add_highlighted_parts_into(
            _produce_variable_name(exception_representation.attribute_name),
            highlighter,
            into,
        )
    
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            ' variable?',
            highlighter,
            into,
        )
        
        into.append('\n')
        lines_rendered += 1
    
    if (not attribute_unset) and (familiar_attribute_names is not None):
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            ('Or perhaps any of the following attributes: ' if lines_rendered >= 2 else 'Did you mean any of: '),
            highlighter,
            into,
        )
        
        index = 0
        length = len(familiar_attribute_names)
        while True:
            name = familiar_attribute_names[index]
            into = add_highlighted_parts_into(_produce_attribute_name(name), highlighter, into)
            
            index += 1
            if index == length:
                break
            
            into = add_highlighted_part_into(
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
                ', ',
                highlighter,
                into,
            )
            continue
        
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            '?',
            highlighter,
            into,
        )
        
        into.append('\n')
        lines_rendered += 1

    if (familiar_attribute_names is None) and  (variable_names_with_attribute is not None):
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            ('Or perhaps you meant to do any of: ' if lines_rendered >= 2 else 'Did you mean to do any of: '),
            highlighter,
            into,
        )
        
        index = 0
        length = len(variable_names_with_attribute)
        while True:
            name = variable_names_with_attribute[index]
            into = add_highlighted_parts_into(
                _produce_variable_attribute_access(name, attribute_name), highlighter, into
            )
            
            index += 1
            if index == length:
                break
            
            into = add_highlighted_part_into(
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
                ', ',
                highlighter,
                into,
            )
            continue
        
        into = add_highlighted_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
            '?',
            highlighter,
            into,
        )
        
        into.append('\n')
        lines_rendered += 1
    
    return into


EXCEPTION_REPRESENTATION_RENDERERS = {
    ExceptionRepresentationAttributeError: _render_exception_representation_attribute_error_into,
    ExceptionRepresentationGeneric: _render_exception_representation_generic_into,
    ExceptionRepresentationSyntaxError: _render_exception_representation_syntax_error_into,
}


def render_exception_representation_into(exception_representation, highlighter, into):
    """
    Renders an exception's representation.
    
    Parameters
    ----------
    exception_representation : `None | ExceptionRepresentationBase`
        The exception representation to render.
    
    highlighter : `None | HighlightFormatterContext`
        Stores how the output should be highlighted.
    
    into : `list<str>`
        The list of strings to extend.
    
    Returns
    -------
    into : `list<str>`
    """
    renderer = EXCEPTION_REPRESENTATION_RENDERERS.get(type(exception_representation), None)
    if (renderer is not None):
        into = renderer(exception_representation, highlighter, into)
    
    return into



import sys
from os import getcwd as get_current_work_directory
from os.path import (
    dirname as get_directory_name, expanduser as get_user_home_directory, join as join_paths,
    normpath as normalize_path, realpath as get_real_path
)


try:
    from . import __package__ as PACKAGE_NAME
except ImportError:
    PACKAGE_NAME = sys.path[0]
    
    sys.path.append(
        normalize_path(
            join_paths(
                get_directory_name(
                    get_real_path(
                        join_paths(
                            get_current_work_directory(),
                            get_user_home_directory(__file__),
                        )
                    )
                ),
                '..',
            )
        )
    )


PACKAGE = __import__(PACKAGE_NAME)
__import__(f'{PACKAGE_NAME}.tools')
__import__(f'{PACKAGE_NAME}.main')

collect_module_variables = PACKAGE.tools.collect_module_variables
run_asynchronous_interactive_console = \
    PACKAGE.tools.asynchronous_interactive_console.run_asynchronous_interactive_console
debug_key = PACKAGE.main.debug_key
get_short_executable = PACKAGE.get_short_executable


def __main__():
    """
    Runs an asynchronous interactive console.
    
    > You need python3.8 or higher to use `await`!
    
    > If `debug-key` or `dk` is passed as a command, it pops up a key debugger instead.
    """
    parameters = sys.argv.copy()
    
    if parameters and ((parameters[0] == __file__) or (parameters[0] != get_short_executable())):
        del parameters[0]
    
    if parameters:
        used_command_name = parameters[0].strip().casefold().replace('_', '-')
    else:
        used_command_name = 'interpreter'
    
    if used_command_name in ('debug-key', 'dk'):
        debug_key()
    
    elif used_command_name in ('interpreter', 'i'):
        run_asynchronous_interactive_console(collect_module_variables(PACKAGE))
    
    else:
        sys.stdout.write(
            'Available commands:\n'
            '- interpreter\n'
            '- debug-key\n'
        )


if __name__ == '__main__':
    __main__()



from .constants import *
from .url import *
from .url_fragment import *
from .url_host_info import *
from .url_part_base import *
from .url_part_base_parsed import *
from .url_path import *
from .url_user_info import *


__all__ = (
    *constants.__all__,
    *url.__all__,
    *url_fragment.__all__,
    *url_host_info.__all__,
    *url_part_base.__all__,
    *url_part_base_parsed.__all__,
    *url_path.__all__,
    *url_user_info.__all__,
)



__all__ = ()

from .expression_parsing import ExpressionInfo
from .frame_group import FrameGroup


def group_frames(frames):
    """
    Groups frames into frame groups and then normalises them into non-repeat and repeat sequences.
    
    Parameters
    ----------
    frames : `list<FrameGroupFrame>`
        The frames to group.
    
    Returns
    -------
    frame_groups : `None | list<FrameGroup>`
        The created frame groups.
    """
    # note: frames may be an empty list.
    if not frames:
        return None
    
    _count_alikes(frames)
    frame_groups = _group_frames_to_frame_groups(frames)
    frame_groups = _merge_frame_groups(frame_groups)
    frame_groups = _separate_repeats_in_frame_groups(frame_groups)
    frame_groups = _merge_frame_groups(frame_groups)
    return frame_groups


def _group_frames_to_frame_groups(frames):
    """
    Groups the frames to frame groups depending how much times a frame's expression is mentioned.
    
    Parameters
    ----------
    frames : `list<FrameGroupFrame>`
        The frames to group.
    
    Returns
    -------
    frame_groups : `list<FrameGroup>`
    """
    frame_group = FrameGroup()
    frame_groups = []
    
    for frame in frames:
        if not frame_group.try_add_frame(frame):
            frame_groups.append(frame_group)
            frame_group = FrameGroup()
            frame_group.try_add_frame(frame)
    
    if frame_group:
        frame_groups.append(frame_group)
    
    return frame_groups


def _merge_frame_groups(frame_groups):
    """
    Merges the given frame groups to produce bigger non-repeat frame groups.
    Call it after creating frame groups to merge too short to be repeated frame groups into single mention groups.
    Call also after separating repeats to merge the cut down non-repeat parts.
    
    Parameters
    ----------
    frame_groups : `list<frameGroup>`
        The frame groups to merge.
    
    Returns
    -------
    frame_groups : `list<FrameGroup>`
    """
    outcome = []
    frame_group_0 = None
    
    for frame_group_1 in frame_groups:
        if (frame_group_0 is not None) and frame_group_0.try_merge(frame_group_1):
            continue
        
        frame_group_0 = frame_group_1
        outcome.append(frame_group_0)
        continue
    
    return outcome


def _separate_repeats_in_frame_groups(frame_groups):
    """
    Separates the repeated parts down. Each frame may produce `itself` or `before - middle (repeated) - after` parts.
    Repeat separating is then again applied to the `before` and `after` parts till every frame group is either
    marked as single-mention or repeat.
    
    Using a flattened algorithm to avoid recursion.
    
    Parameters
    ----------
    frame_groups : `list<frameGroup>`
        The frame groups to merge.
    
    Returns
    -------
    frame_groups : `list<FrameGroup>`
    """
    frame_groups_to_do = [(False, frame_group) for frame_group in reversed(frame_groups)]
    frame_groups = []
    
    while frame_groups_to_do:
        checked, frame_group = frame_groups_to_do.pop()
        if checked:
            frame_groups.append(frame_group)
        else:
            # cant assign it in order normally, so we reverse add it
            frame_groups_to_do.extend(reversed([*frame_group.iter_separate_repeated()]))
    
    return frame_groups


def _count_alikes(frames):
    """
    Counts how much alike frames there are.
    
    Parameters
    ----------
    frames : `list<FrameProxyBase>`
        A list of frame proxies to create expression info for.
    """
    # Collect
    expressions = {}
    
    for frame in frames:
        expression_key = frame.expression_key
        
        try:
            alike_frames = expressions[expression_key]
        except KeyError:
            alike_frames = []
            expressions[expression_key] = alike_frames
        
        alike_frames.append(frame)
    
    # count
    for alike_frames in expressions.values():
        alike_count = len(alike_frames)
        for frame in alike_frames:
            frame.alike_count = alike_count


def normalize_frame_groups(frame_groups):
    """
    Normalises the given frame groups removing the empty ones.
    
    Parameters
    ----------
    frame_groups : `None | list<frameGroup>`
        The frame groups to merge.
    
    Returns
    -------
    frame_groups : `None | list<FrameGroup>`
    """
    if frame_groups is None:
        return None
    
    outcome = None
    
    for frame_group in frame_groups:
        if not frame_group:
            continue
        
        if outcome is None:
            outcome = []
        
        outcome.append(frame_group)
    
    return outcome



from .client_request import *
from .client_response import *
from .connection import *
from .connection_key import *
from .connector_base import *
from .connector_tcp import *
from .constants import *
from .host_info import *
from .host_info_basket import *
from .http_client import *
from .request_context_manager import *
from .request_info import *
from .protocol_basket import *  
from .proxy import *
from .ssl_fingerprint import *
from .web_socket_context_manager import *


__all__ = (
    *client_request.__all__,
    *client_response.__all__,
    *connection.__all__,
    *connection_key.__all__,
    *connector_base.__all__,
    *connector_tcp.__all__,
    *constants.__all__,
    *host_info.__all__,
    *host_info_basket.__all__,
    *http_client.__all__,
    *request_context_manager.__all__,
    *protocol_basket.__all__,
    *proxy.__all__,
    *request_info.__all__,
    *ssl_fingerprint.__all__,
    *web_socket_context_manager.__all__,
)


# Deprecations

from warnings import warn

def __getattr__(attribute_name):
    if attribute_name == 'TCPConnector':
        warn(
            (
                f'`TCPConnector` has been renamed to `ConnectorTCP`.'
                f'`ConnectorTCP` will be removed in 2025 August.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return ConnectorTCP
    
    raise AttributeError(attribute_name)



__all__ = ()

from ..utils import RichAttributeErrorBaseType

from .constants import (
    KEEP_ALIVE_CONNECTION_TIMEOUT_DEFAULT, KEEP_ALIVE_CONNECTION_TIMEOUT_KEY, KEEP_ALIVE_HEADER_RP,
    KEEP_ALIVE_MAX_REQUESTS_DEFAULT, KEEP_ALIVE_MAX_REQUESTS_KEY
)


class KeepAliveInfo(RichAttributeErrorBaseType):
    """
    Keep-alive information of a connection.
    
    Attributes
    ----------
    connection_timeout : `float`
        The connection's timeout.
    
    max_requests : `int`
        The amount of maximal allowed requests.
    """
    __slots__ = ('connection_timeout', 'max_requests')
    
    def __new__(cls, connection_timeout, max_requests):
        """
        Creates a new keep-alive info.
        
        Parameters
        ----------
        connection_timeout : `float`
            The connection's timeout.
        
        max_requests : `int`
            The amount of maximal allowed requests.
        """
        self = object.__new__(cls)
        self.connection_timeout = connection_timeout
        self.max_requests = max_requests
        return self
    
    
    def __repr__(self):
        """Returns the keep-alive's representation"""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # connection_timeout
        connection_timeout = self.connection_timeout
        if connection_timeout:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' connection_timeout = ')
            repr_parts.append(repr(connection_timeout))
        
        # max_requests
        max_requests = self.max_requests
        if max_requests != KEEP_ALIVE_CONNECTION_TIMEOUT_DEFAULT:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' max_requests = ')
            repr_parts.append(repr(max_requests))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two keep-alives are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # connection_timeout
        if self.connection_timeout != other.connection_timeout:
            return False
        
        if self.max_requests != other.max_requests:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the keep-alive's hash value."""
        hash_value = 0
        
        # connection_timeout
        connection_timeout = self.connection_timeout
        if connection_timeout != KEEP_ALIVE_CONNECTION_TIMEOUT_DEFAULT:
            hash_value ^= hash(connection_timeout)
        
        # max_requests
        max_requests = self.max_requests
        if max_requests != KEEP_ALIVE_MAX_REQUESTS_DEFAULT:
            hash_value ^= max_requests << 8
        
        return hash_value
    
    
    @classmethod
    def create_default(cls):
        """
        Creates a default keep-alive information.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.connection_timeout = KEEP_ALIVE_CONNECTION_TIMEOUT_DEFAULT
        self.max_requests = KEEP_ALIVE_MAX_REQUESTS_DEFAULT
        return self
    
    
    @classmethod
    def from_header_value(cls, header_value):
        """
        Creates a new keep-alive information from the given header value.
        
        Parameters
        ----------
        header_value : `None | str`
            Header value.
        
        Returns
        -------
        self : `instance<cls>`
        """
        connection_timeout = KEEP_ALIVE_CONNECTION_TIMEOUT_DEFAULT
        max_requests = KEEP_ALIVE_MAX_REQUESTS_DEFAULT
        
        if (header_value is not None):
            for match in KEEP_ALIVE_HEADER_RP.finditer(header_value):
                key, value = match.groups()
                key = key.casefold()
                
                if key == KEEP_ALIVE_CONNECTION_TIMEOUT_KEY:
                    try:
                        connection_timeout = float(value)
                    except ValueError:
                        pass
                    
                    continue
                
                if key == KEEP_ALIVE_MAX_REQUESTS_KEY:
                    try:
                        max_requests = int(value)
                    except ValueError:
                        pass
                    
                    continue
                
                # unknown key, ignore it.
        
        self = object.__new__(cls)
        self.connection_timeout = connection_timeout
        self.max_requests = max_requests
        return self
    
    
    def to_header_value(self):
        """
        Converts the keepa-live information to its header value representation.
        
        Returns
        -------
        header_value : `None | str`
        """
        output_parts = None
        
        # connection_timeout
        connection_timeout = self.connection_timeout
        if connection_timeout != KEEP_ALIVE_CONNECTION_TIMEOUT_DEFAULT:
            if output_parts is None:
                output_parts = []
            
            output_parts.append(KEEP_ALIVE_CONNECTION_TIMEOUT_KEY)
            output_parts.append('=')
            output_parts.append(format(connection_timeout, '.0f'))
        
        # max_requests
        max_requests = self.max_requests
        if max_requests != KEEP_ALIVE_MAX_REQUESTS_DEFAULT:
            if output_parts is None:
                output_parts = []
            else:
                output_parts.append(',')
            
            output_parts.append(KEEP_ALIVE_MAX_REQUESTS_KEY)
            output_parts.append('=')
            output_parts.append(str(max_requests))
        
        # Build
        if (output_parts is not None):
            return ''.join(output_parts)



__all__ = ('ExceptionRepresentationSyntaxError',)

from ...docs import copy_docs

from .exception_representation_base import ExceptionRepresentationBase
from .representation_helpers import _get_type_name


class ExceptionRepresentationSyntaxError(ExceptionRepresentationBase):
    """
    Holds a valid syntax error's representation.
    
    Attributes
    ----------
    file_name : `str`
        The file's name where the exception occurred.
    line : `str`
        The line where the syntax is invalid. Can be empty string.
    line_index : `int`
        The line's index in the file.
    message : `str`
        Error message. Can not be empty string.
    pointer_start_offset : `int`
        The offset after the pointer starts.
    pointer_length : `int`
        The pointer's length. Can be `0` if it should not be shown.
    type_name : `str`
        The represented exception's type's name.
    """
    __slots__ = ('file_name', 'line', 'line_index', 'message', 'pointer_start_offset', 'pointer_length', 'type_name')
    
    def __new__(cls, exception, frames):
        """
        Creates a new exception representation.
        
        Parameters
        ----------
        exception : `SyntaxError`
            Exception to represent.
        frames : `None | list<FrameProxyBase>`
            The frames the exception is raised with.
        """
        message, frames = exception.args
        
        if (message is None):
            message = ''
        
        file_name = frames[0]
        line_index = frames[1] - 1
        pointer_start_offset = frames[2] - 1
        line = frames[3]
        
        if len(frames) == 6:
            pointer_end_offset = frames[5] - 1
        else:
            pointer_end_offset = pointer_start_offset + 1
        
        if (file_name is None):
            file_name = ''
        
        if (line is None):
            line = ''
            pointer_start_offset = 0
            pointer_length = 0
            
        else:
            line = line.rstrip()
            
            left_stripped_count = 0
            for character in line:
                if character in {' ', '\n', '\t', '\f'}:
                    left_stripped_count += 1
                    continue
                
                break
            
            line = line[left_stripped_count:]
            
            pointer_start_offset -= left_stripped_count
            pointer_end_offset -= left_stripped_count
        
            if pointer_start_offset < 0:
                pointer_start_offset = 0
            
            if pointer_end_offset < 0:
                pointer_end_offset = 0
            
            if pointer_start_offset >= pointer_end_offset:
                pointer_start_offset = 0
                pointer_length = 0
            else:
                pointer_length = pointer_end_offset - pointer_start_offset
        
        type_name = _get_type_name(type(exception))
        
        self = object.__new__(cls)
        self.file_name = file_name
        self.line = line
        self.line_index = line_index
        self.message = message
        self.pointer_length = pointer_length
        self.pointer_start_offset = pointer_start_offset
        self.type_name = type_name
        return self
    
    
    @classmethod
    def from_fields(
        cls,
        *,
        file_name = ...,
        line = ...,
        line_index = ...,
        message = ...,
        pointer_start_offset = ...,
        pointer_length = ...,
        type_name = ...,
    ):
        """
        Creates a new syntax error representation from the given fields.
        
        Attributes
        ----------
        file_name : `str`, Optional (Keyword only)
            The file's name where the exception occurred.
        line : `str`, Optional (Keyword only)
            The line where the syntax is invalid. Can be empty string.
        line_index : `int`, Optional (Keyword only)
            The line's index in the file.
        message : `str`, Optional (Keyword only)
            Error message. Can not be empty string.
        pointer_start_offset : `int`, Optional (Keyword only)
            The offset after the pointer starts.
        pointer_length : `int`, Optional (Keyword only)
            The pointer's length. Can be `0` if it should not be shown.
        type_name : `str`, Optional (Keyword only)
            The represented exception's type's name.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.file_name = '' if file_name is ... else file_name
        self.line = '' if line is ... else line
        self.line_index = 0 if line_index is ... else line_index
        self.message = '' if message is ... else message
        self.pointer_start_offset = 0 if pointer_start_offset is ... else pointer_start_offset
        self.pointer_length = 0 if pointer_length is ... else pointer_length
        self.type_name = SyntaxError.__name__ if type_name is ... else type_name
        return self
    
    
    @copy_docs(ExceptionRepresentationBase._populate_repr_parts)
    def _populate_repr_parts(self, repr_parts):
        # type_name
        repr_parts.append(' type_name = ')
        repr_parts.append(repr(self.type_name))
        
        # message
        message = self.message
        if (message is not None):
            repr_parts.append(', message = ')
            repr_parts.append(repr(message))
        
        # location
        file_name = self.file_name
        line_index = self.line_index
        if file_name and line_index >= 0:
            repr_parts.append(', file_name = ')
            repr_parts.append(repr(file_name))
            repr_parts.append(', line_index = ')
            repr_parts.append(repr(self.line_index))
        
        
        # line & pointer
        line = self.line
        if line:
            repr_parts.append(', line = ')
            repr_parts.append(repr(line))
            
            pointer_length = self.pointer_length
            if pointer_length:
                repr_parts.append(', pointer_start_offset = ')
                repr_parts.append(repr(self.pointer_start_offset))
                
                repr_parts.append(', pointer_length = ')
                repr_parts.append(repr(pointer_length))
    
    
    @copy_docs(ExceptionRepresentationBase._is_equal)
    def _is_equal(self, other):
        if self.type_name != other.type_name:
            return False
        
        if self.message != other.message:
            return False
        
        if self.file_name != other.file_name:
            return False
        
        if self.line_index != other.line_index:
            return False
        
        if self.line != other.line:
            return False
        
        if self.pointer_start_offset != other.pointer_start_offset:
            return False
        
        if self.pointer_length != other.pointer_length:
            return False
        
        return True



__all__ = ('WebSocketFrame', )

from warnings import warn

from ..utils import RichAttributeErrorBaseType

from .exceptions import WebSocketProtocolError


WEB_SOCKET_OPERATION_CONTINUOUS = 0
WEB_SOCKET_OPERATION_TEXT = 1
WEB_SOCKET_OPERATION_BINARY = 2

WEB_SOCKET_OPERATION_CLOSE = 8
WEB_SOCKET_OPERATION_PING = 9
WEB_SOCKET_OPERATION_PONG = 10

WEB_SOCKET_DATA_OPERATIONS = (WEB_SOCKET_OPERATION_CONTINUOUS,  WEB_SOCKET_OPERATION_TEXT, WEB_SOCKET_OPERATION_BINARY)
WEB_SOCKET_CONTROL_OPERATIONS = (WEB_SOCKET_OPERATION_CLOSE, WEB_SOCKET_OPERATION_PING, WEB_SOCKET_OPERATION_PONG)

# TODO: whats the fastest way on pypy ? casting 64 bit ints -> xor -> replace?
# You can do `data = memoryview(data).cast('Q', (size,))` and then index max, but that is still 3 times slower
# I use pypy3.6 tho, so perhaps on newer versions it is faster

_XOR_TABLE = [bytes(a ^ b for a in range(256)) for b in range(256)]


def apply_web_socket_mask(mask, data):
    """
    Applies web socket mask on the given data and returns a new instance.
    
    Parameters
    ----------
    data : `bytes-like`
        Data to apply the mask to.
    mask : `bytes`
        `uint32` web socket mask in bytes.
    """
    data_bytes = bytearray(data)
    for index in range(4):
        data_bytes[index::4] = data_bytes[index::4].translate(_XOR_TABLE[mask[index]])
    return data_bytes


class WebSocketFrame(RichAttributeErrorBaseType):
    """
    Represents a web socket frame.
    
    Attributes
    ----------
    data : `bytes`
        The data of the frame.
    
    head_0 : `int`
        The first byte of the web socket frame.
    """
    __slots__ = ('data', 'head_0')
    
    def __new__(cls, final, operation_code, data):
        """
        Creates a web socket frame.
        
        Parameters
        -----------
        final : `bool`
            Whether this web socket frame is a final web socket frame.
            When sending web socket frames, the data of frames it collected, till a final frame is received.
        
        operation_code : `int`
            The operation code of the web socket frame.
            
            Can be any of the following:
            
            +-----------------------------------+-------+
            | Respective name                   | Value |
            +===================================+=======+
            | WEB_SOCKET_OPERATION_CONTINUOUS   | 0     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_TEXT         | 1     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_BINARY       | 2     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_CLOSE        | 8     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_PING         | 9     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_PONG         | 10    |
            +-----------------------------------+-------+
        
        data : `bytes`
            The data to ship with the web socket frame.
        """
        self = object.__new__(cls)
        self.data = data
        self.head_0 = (final << 7) | operation_code
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two web socket frames are equal."""
        if type(self) is not type(self):
            return NotImplemented
        
        if self.head_0 != other.head_0:
            return False
        
        if self.data != other.data:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the websocket frame's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # final
        repr_parts.append(' final = ')
        repr_parts.append(repr(self.final))
        
        repr_parts.append(', operation_code = ')
        repr_parts.append(repr(self.operation_code))
        
        repr_parts.append(', data (length) = ')
        repr_parts.append(repr(len(self.data)))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
        
    
    @property
    def final(self):
        """
        Returns whether the web socket frame is final.
        
        Returns
        -------
        is_final : `bool`
        """
        return True if (self.head_0 & 0b10000000) else False
    
    
    @property
    def rsv1(self):
        """
        Returns the first reserved bit of the web socket frame's head.
        
        Returns
        -------
        rsv1 : `int`
        """
        warn(
            f'`{type(self).__name__}.rsv1` is deprecated and will be removed in 2025 October.',
            FutureWarning,
            stacklevel = 2,
        )
        
        return True if (self.head_0 & 0b01000000) else False
    
    
    @property
    def rsv2(self):
        """
        Returns the second reserved bit of the web socket frame's head.
        
        Returns
        -------
        rsv2 : `int`
        """
        warn(
            f'`{type(self).__name__}.rsv2` is deprecated and will be removed in 2025 October.',
            FutureWarning,
            stacklevel = 2,
        )
        
        return True if (self.head_0 & 0b00100000) else False
    
    
    @property
    def rsv3(self):
        """
        Returns the third reserved bit of the web socket frame's head.
        
        Returns
        -------
        rsv3 : `int`
        """
        warn(
            f'`{type(self).__name__}.rsv3` is deprecated and will be removed in 2025 October.',
            FutureWarning,
            stacklevel = 2,
        )
        
        return True if (self.head_0 & 0b00010000) else False
    
    
    @property
    def operation_code(self):
        """
        Returns the web socket frame's operation_code.
        
        Returns
        -------
        operation_code : `int`
            Can be one of the following values:
            
            +-----------------------------------+-------+
            | Respective name                   | Value |
            +===================================+=======+
            | WEB_SOCKET_OPERATION_CONTINUOUS   | 0     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_TEXT         | 1     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_BINARY       | 2     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_CLOSE        | 8     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_PING         | 9     |
            +-----------------------------------+-------+
            | WEB_SOCKET_OPERATION_PONG         | 10    |
            +-----------------------------------+-------+
        """
        return self.head_0 & 0b00001111
    
    
    def check(self):
        """
        Check that this frame contains acceptable values.
        
        Raises
        ------
        WebSocketProtocolError
            - If the reserved bits are not `0`.
            - If the frame is a control frame, but is too long for one.
            - If the web socket frame is fragmented frame. (Might be supported if people request is.)
            - If the frame operation_code is not any of the expected ones.
        """
        if self.head_0 & 0b01110000:
            raise WebSocketProtocolError('Reserved bits must be `0`.')
        
        operation_code = self.head_0 & 0b00001111
        if operation_code in WEB_SOCKET_DATA_OPERATIONS:
            return
        
        if operation_code in WEB_SOCKET_CONTROL_OPERATIONS:
            if len(self.data) > 125:
                raise WebSocketProtocolError('Control frame too long.')
            if not self.head_0 & 0b10000000:
                raise WebSocketProtocolError('Fragmented control frame.')
            return
        
        raise WebSocketProtocolError(f'Invalid operation_code: {operation_code}.')
    
    
    @classmethod
    def _from_fields(cls, head_0, data):
        """
        Creates a new web socket frame instance.
        
        Parameters
        ----------
        head_0 : `int`
            The 0th byte of web socket frame. It holds all the required data.
        
        data : `bytes`
            The data of the frame.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.head_0 = head_0
        self.data = data
        return self
    
    
    @property
    def head_1(self):
        """
        Deprecated and will be removed in 2025 october.
        """
        warn(
            (
                f'`{type(self).__name__}.head_1` is deprecated and will be removed in 2025 October. '
                f'Please use `.head_0` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.head_0
    
    
    @property
    def is_final(self):
        """
        Deprecated and will be removed in 2025 october.
        """
        warn(
            (
                f'`{type(self).__name__}.is_final` is deprecated and will be removed in 2025 October. '
                f'Please use `.final` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.final



__all__ = ()

from ....multi_value_dictionary import MultiValueDictionary

from difflib import SequenceMatcher
from heapq import nlargest as get_n_largest
from itertools import permutations as permutate


SUGGESTIONS_MAX = 25


def _split_part_by_numbers(part):
    """
    Splits the given string part by numbers.
    
    This function is an iterable generator.
    
    Examples
    --------
    ```py
    325 -> 325
    koishi -> koishi
    koishi3 -> koishi, 3
    3koishi -> 3, koishi
    46koishi12satori -> 46, satori, 12, satori
    ```
    
    Parameters
    ----------
    part : `str`
        String part.
    
    Yields
    ------
    part : `str`
    """
    previous_characters = []
    
    for character in part:
        if previous_characters and previous_characters[-1].isnumeric() != character.isnumeric():
            yield ''.join(previous_characters)
            previous_characters.clear()
        
        previous_characters.append(character)
        continue
    
    if previous_characters:
        yield ''.join(previous_characters)


def _split_part_by_casing(part):
    """
    Splits the given string parts by casing.
    
    This function is an iterable generator.
    
    Examples
    --------
    ```py
    koishi -> koishi
    KOISHI -> KOISHI
    Koishi -> Koishi
    koishiSatori -> koishi, Satori
    KOISHISatori -> KOISHI, Satori
    koishiSATORI -> koishi, SATORI
    KOISHIsatori -> KOISH, Isatori # There is a limit how much we can do
    ```
    
    Parameters
    ----------
    part : `str`
        String part.
    
    Yields
    ------
    part : `str`
    """
    previous_characters = []
    
    for character in part:
        if previous_characters:
            if character.isupper():
                if not previous_characters[-1].isupper():
                    yield ''.join(previous_characters)
                    previous_characters.clear()
            
            else:
                if previous_characters[-1].isupper() and len(previous_characters) > 1:
                    yield ''.join(previous_characters[:-1])
                    del previous_characters[:-1]
        
        previous_characters.append(character)
        continue
    
    if previous_characters:
        yield ''.join(previous_characters)


def _iter_split_name_to_words(name):
    """
    Splits the given name to words.
    
    This function is an iterable generator.
    
    Examples
    --------
    ```py
    koishi_satori_69 -> koishi, satori, 69
    KoishiSatori69 -> Koishi, Satori, 69
    KOISHI_SATORI_69 -> KOISHI, Satori, 69
    KOISHISATORI69 -> KOISHISATORI, 69 # there is a limit how much we can do
    ```
    
    Parameters
    ----------
    name : `str`
        The name to split.
    
    Yields
    ------
    word : `str`
    """
    for part in name.split('_'):
        if part:
            for part in _split_part_by_numbers(part):
                yield from _split_part_by_casing(part)


def _normalize_name(name):
    """
    Normalize the given name.
    
    Parameters
    ----------
    name : `str`
        The name to normalize.
    
    Returns
    -------
    name : `str`
    """
    return '_'.join([word.casefold() for word in _iter_split_name_to_words(name)])


def _iter_name_aliases(name):
    """
    Alters the given `name`.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    name : `str`
        The name to alter.
    
    Yields
    ------
    alternative_name : `str`
    """
    words = [*_iter_split_name_to_words(name)]
    for permutation in permutate(words, len(words)):
        yield '_'.join([word.casefold() for word in permutation])


def _get_familiar_names(names, name):
    """
    Gets suggestions for the given name.
    
    Parameters
    ----------
    names : `set<str>`
        Names to get suggestions from.
    name : `str`
        Name to get suggestions for.
    
    Returns
    -------
    suggestions : `list<str>`
    """
    name_aliases = [*_iter_name_aliases(name)]
    names_map = MultiValueDictionary((_normalize_name(name), name) for name in names)
    
    sequence_matcher = SequenceMatcher()
    
    matches = {}
    
    for name in name_aliases:
        required_match_level = 0.8 - (20 - min(len(name), 20)) * 0.01
        sequence_matcher.set_seq2(name)
        
        for name in names_map.keys():
            sequence_matcher.set_seq1(name)
            if sequence_matcher.real_quick_ratio() < required_match_level:
                continue
            
            if sequence_matcher.quick_ratio() < required_match_level:
                continue
            
            ratio = sequence_matcher.ratio()
            if ratio < required_match_level:
                continue
            
            # If the same name shows up twice, use the highest available ratio.
            matches[name] = max(ratio, matches.get(name, 0.0))
    
    # Pull suggestions
    suggestions = []
    for match in get_n_largest(SUGGESTIONS_MAX, [(ratio, name) for name, ratio in matches.items()]):
        matches = names_map.get_all(match[1])
        matches.sort()
        suggestions.extend(matches)
    
    return suggestions



__all__ = ()

from ...utils import copy_docs

from .url_part_base import URLPartBase


class URLHostPart(URLPartBase):
    """
    Represents a url's host's part..
    
    Attributes
    ----------
    _decoded : `None | str`
        Decoded value.
    
    _encoded : `None | str`
        Encoded value.
    
    _flags : `int`
        Bitwise flags representing how the object is constructed.
    """
    __slots__ = ()
    
    @staticmethod
    @copy_docs(URLPartBase._encode)
    def _encode(value):
        if (value is not None):
            try:
                value.encode('ascii')
            except UnicodeEncodeError:
                value = value.encode('idna').decode('ascii')
            
        return value
    
    
    @staticmethod
    @copy_docs(URLPartBase._decode)
    def _decode(value):
        if (value is not None):
            try:
                value = value.encode('ascii').decode('idna')
            except UnicodeEncodeError:
                pass
        
        return value



__all__ = ('ExceptionProxyBase',)

from ..exception_representation import ExceptionRepresentationBase
from ..frame_grouping import normalize_frame_groups


class ExceptionProxyBase:
    """
    Base type for exception proxies.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new exception proxy.
        """
        return object.__new__(cls)
    
    
    @property
    def exception_representation(self):
        """
        Returns the exception's representation.
        
        Returns
        -------
        exception_representation : `None | ExceptionRepresentationBase`
        """
        return None
    
    
    @property
    def frame_groups(self):
        """
        Returns the exception's grouped frames.
        
        Returns
        -------
        frame_groups : `None | list<FrameGroup>`
        """
        return None
    
    
    @frame_groups.setter
    def frame_groups(self, value):
        if (value is not None):
            raise RuntimeError('Cannot set attribute to non-default.')
    
    
    def __repr__(self):
        """Returns the exception proxy's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # exception_representation
        exception_representation = self.exception_representation
        if (exception_representation is not None):
            repr_parts.append(' exception_representation = ')
            repr_parts.append(repr(exception_representation))
            
            field_added = True
        else:
            field_added = False
        
        # frame_groups
        frame_groups = self.frame_groups
        if (frame_groups is not None):
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' frame_groups = ')
            repr_parts.append(repr(frame_groups))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two exception proxies are equal."""
        if not isinstance(other, ExceptionProxyBase):
            return NotImplemented
        
        return self._is_equal(other)
    
    
    def __len__(self):
        """Returns the amount of frames in the exception proxy."""
        return sum(len(frame_group) for frame_group in self.iter_frame_groups())
    
    
    def _is_equal(self, other):
        """
        Returns whether the two exception proxies are equal.
        
        Parameters
        ----------
        other : ``ExceptionProxyBase``
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        if self.exception_representation != other.exception_representation:
            return False
        
        if self.frame_groups != other.frame_groups:
            return False
        
        return True
    
    
    def __mod__(self, other):
        """Returns whether the two exception proxies are alike."""
        if not isinstance(other, ExceptionProxyBase):
            return NotImplemented
        
        return self._is_alike(other)
    
    
    def __rmod__(self, other):
        """Returns whether the two exception proxies are alike."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_alike(other)
    
    
    def _is_alike(self, other):
        """
        Returns whether the exception proxies are alike.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_alike : `bool`
        """
        # exception_representation
        if self.exception_representation != other.exception_representation:
            return False
        
        # frame_groups
        if len(self) != len(other):
            return False
        
        for self_frame_group, other_frame_group in zip(self.iter_frame_groups(), other.iter_frame_groups()):
            if not (self_frame_group % other_frame_group):
                return False
        
        return True
    
    
    def drop_variables(self):
        """
        Drops the variables of the exception representation.
        """
        for frame_group in self.iter_frame_groups():
            frame_group.drop_variables()
    
    
    def has_variables(self):
        """
        Returns whether any frame group inherited variables on creation.
        
        Returns
        -------
        has_variables : `bool`
        """
        for frame_group in self.iter_frame_groups():
            if frame_group.has_variables():
                return True
        
        return False
    
    
    def iter_frame_groups(self):
        """
        Iterates over the frame groups of the exception proxy.
        
        This method is an iterable generator.
        
        Yields
        ------
        frame_group : ``FrameGroup``
        """
        frame_groups = self.frame_groups
        if (frame_groups is not None):
            yield from frame_groups
    
    
    def drop_ignored_frames(self, *, filter = None):
        """
        Drops the frames that should be ignored.
        
        Parameters
        ----------
        filter : `None | callable` = `None`, Optional (Keyword only)
            Additional filter to check whether a frame should be shown.
        """
        for frame_group in self.iter_frame_groups():
            frame_group.drop_ignored_frames(filter = filter)
        
        self.frame_groups = normalize_frame_groups(self.frame_groups)
    
    
    def apply_frame_filter(self, filter):
        """
        Keeps the frames that pass the given `filter`.
        
        Parameters
        ----------
        filter : `callable`
            Filter to check whether a frame should be kept.
        """
        for frame_group in self.iter_frame_groups():
            frame_group.apply_frame_filter(filter)
        
        self.frame_groups = normalize_frame_groups(self.frame_groups)
    
    
    def iter_frames_no_repeat(self):
        """
        Iterates over the frames of the exception proxy without repeating repeated blocks.
        It may still happen that non-unique frames are yielded.
        
        This method is an iterable generator.
        
        Yields
        ------
        frame : ``FrameProxyBase``
        """
        for frame_group in self.iter_frame_groups():
            yield from frame_group.iter_frames_no_repeat()



__all__ = ()


def _get_variable_names(frame):
    """
    Gets the variable names from the given frame proxy for suggestions.
    
    Parameters
    ----------
    frame : `None | FrameProxyBase`
        Frame to get variable names from.
    
    Returns
    -------
    variable_names : `set<str>`
    """
    variable_names = set()
    
    if (frame is not None):
        locals = frame.locals
        if (locals is not None):
            variable_names.update(locals.keys())
        
        globals = frame.globals
        if (globals is not None):
            variable_names.update(globals.keys())
        
        builtins = frame.builtins
        if (builtins is not None):
            variable_names.update(builtins.keys())
        
    return variable_names


def _iter_variables(frame_proxy):
    """
    Iterates over the variables of the frame profy for suggestions.
    
    Parameters
    ----------
    frame_proxy : `None | FrameProxyBase`
        Frame proxy to get variables of.
    
    Yields
    ------
    variable_name : `str`
    variable_value : `object`
    """
    if (frame_proxy is not None):
        locals = frame_proxy.locals
        if (locals is not None):
            yield from locals.items()
        
        globals = frame_proxy.globals
        if (globals is not None):
            yield from globals.items()
        
        builtins = frame_proxy.builtins
        if (builtins is not None):
            yield from builtins.items()


def _iter_unique_variables(frame_proxy):
    """
    Iterates over the variables of the frame profy for suggestions.
    If a variable name is present twice, only yields it the first time.
    
    Parameters
    ----------
    frame_proxy : `None | FrameProxyBase`
        Frame proxy to get variables of.
    
    Yields
    ------
    variable_name : `str`
    variable_value : `object`
    """
    length = 0
    collected = set()
    
    for variable_name, variable_value in  _iter_variables(frame_proxy):
        collected.add(variable_name)
        new_length = len(collected)
        if length != new_length:
            yield variable_name, variable_value
        
        length = new_length



__all__ = ('HttpReadProtocol', 'HttpReadWriteProtocol',)

from base64 import b64decode as base64_decode
from binascii import a2b_qp as qp_decode
from functools import partial as partial_func
from random import getrandbits
from struct import Struct

from ..core import ReadProtocolBase, ReadWriteProtocolBase
from ..utils import IgnoreCaseMultiValueDictionary, copy_docs

from .compressors import COMPRESSION_ERRORS, get_decompressor_for
from .exceptions import PayloadError, WebSocketProtocolError
from .headers import CONTENT_ENCODING, CONTENT_LENGTH, CONTENT_TRANSFER_ENCODING, CONTENT_TYPE, METHOD_CONNECT
from .helpers import HttpVersion11, parse_http_headers, parse_http_request, parse_http_response
from .http_message import RawRequestMessage, RawResponseMessage
from .content_type import parse_content_type
from .web_socket_frame import WebSocketFrame, apply_web_socket_mask


PAYLOAD_ERROR_EOF_AT_HTTP_HEADER = 'EOF received meanwhile reading http headers.'


UNPACK_LENGTH_2 = Struct('!H').unpack_from
UNPACK_LENGTH_3 = Struct('!Q').unpack_from

PACK_LENGTH_1 = Struct('!BB').pack
PACK_LENGTH_2 = Struct('!BBH').pack
PACK_LENGTH_3 = Struct('!BBQ').pack


class HttpReadProtocol(ReadProtocolBase):
    """
    Asynchronous read protocol implementation. Stuffed full of http read methods.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    
    _chunks : `Deque<bytes>`
        Right feed, left pop queue, used to store the received data chunks.
    
    _exception : `None | BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    
    _payload_reader : `None | GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    
    _payload_stream : `None | PayloadStream`
        Payload stream of the protocol.
    
    _transport : `None | AbstractTransportLayerBase`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    """
    __slots__ = ()
    
    
    def should_close(self):
        """
        Returns the protocol should be closed.
        
        Returns
        -------
        should_close : `bool`
        """
        if (self._payload_stream is not None):
            return True
        
        if (self._exception is not None):
            return True
        
        if (not self._at_eof) and (self._offset or self._chunks):
            return True
        
        return False
    
    
    async def read_http_response(self):
        """
        Reads http response.
        
        This method is a coroutine.
        
        Returns
        -------
        response_message : ``RawResponseMessage``
        
        Raises
        ------
        PayloadError
            Invalid data received.
        """
        try:
            data = await self.read_until(b'\r\n\r\n')
        except ConnectionError as exception:
            cause = exception.__cause__
            if (cause is not None) and isinstance(cause, EOFError):
                raise PayloadError(PAYLOAD_ERROR_EOF_AT_HTTP_HEADER) from exception
            
            raise
        
        return parse_http_response(data)
    
    
    async def read_http_request(self):
        """
        Reads http request.
        
        This method is a coroutine.
        
        Returns
        -------
        request_message : ``RawResponseMessage``
        
        Raises
        ------
        EOFError
            Connection lost before enough data was received.
        PayloadError
            Invalid data received.
        """
        try:
            data = await self.read_until(b'\r\n\r\n')
        except ConnectionError as exception:
            cause = exception.__cause__
            if (cause is not None) and isinstance(cause, EOFError):
                raise PayloadError(PAYLOAD_ERROR_EOF_AT_HTTP_HEADER) from exception
            
            raise
        
        return parse_http_request(data)
    
    
    async def _read_single_multipart(self, boundary, is_first):
        """
        Reads a single multipart.
        
        This method is a coroutine.
        
        Parameters
        ----------
        boundary : `bytes`
            A boundary that marks the start and the end of the multipart.
        
        is_first : `bool`
            Whether this is the first multipart that we read.
        
        Returns
        -------
        is_more : `bool`
            Whether the payload contains more multipart field.
        
        headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`)
            Received response headers.
        
        chunk : `None | bytes`
            The field content.
        
        Raises
        ------
        EOFError
            Connection lost before enough data was received.
        PayloadError
            Invalid data received.
        """
        if is_first:
            await self.read_until(b'--' + boundary)
            
            try:
                maybe_end_0 = await self.read_exactly(2)
            except EOFError:
                # End of payload? Ok i guess.
                return False, None, None
            else:
                if maybe_end_0 == b'\r\n':
                    pass
                elif maybe_end_0 == b'--':
                    # End of payload?
                    try:
                        maybe_end_1 = await self.read_exactly(2)
                    except EOFError:
                        return False, None, None
                    else:
                        if maybe_end_1 == b'\r\n':
                            return False, None, None
                        else:
                            raise PayloadError(
                                f'Multipart boundary not ended with b\'--\' + b\'\r\n\', got '
                                f'b\'--\'+{maybe_end_1!r}.'
                            )
                else:
                    raise PayloadError(
                        f'Multipart boundary not ended either with b\'--\' or b\'\r\n\', got '
                        f'{maybe_end_0!r}.'
                    )
        
        data = await self.read_until(b'\r\n\r\n')
        headers = parse_http_headers(data, 0)
        
        length = headers.get(CONTENT_LENGTH, None)
        if length is None:
            part = await self.read_until(b'\r\n--' + boundary)
        else:
            length = int(length)
            part = await self._read_exactly_as_one(length)
            try:
                maybe_boundary = await self.read_exactly(len(boundary) + 4)
            except EOFError:
                return False, None, part
            
            if maybe_boundary != b'\r\n--' + boundary:
                raise PayloadError(
                    f'Multipart payload not ended with boundary, expected: b\'\r\n\' + b\'--\' + '
                    f'{boundary!r}, got {maybe_boundary!r}.'
                )
            
            
        try:
            maybe_end_0 = await self.read_exactly(2)
        except EOFError:
            return False, headers, part
        
        if maybe_end_0 == b'\r\n':
            return True, headers, part
        
        if maybe_end_0 == b'--':
            try:
                maybe_end_1 = await self.read_exactly(2)
            except EOFError:
                return False, headers, part
            
            if maybe_end_1 == b'\r\n':
                return False, headers, part
            
            raise PayloadError(
                f'Multipart boundary not ended with b\'--\'+b\'\r\n\', got '
                f'b\'--\'+{maybe_end_1!r}.'
            )
        
        raise PayloadError(
            f'Multipart boundary not ended either with  b\'--\' or b\'\r\n\', got '
            f'{maybe_end_0!r}.'
        )
    
    
    async def read_multipart(self, headers):
        """
        Reads multipart data from the protocol.
        
        This method is a coroutine generator.
        
        Parameters
        ----------
        headers : ``IgnoreCaseMultiValueDictionary``
            The response's or the request's headers.
        
        Yields
        ------
        headers : `IgnoreCaseMultiValueDictionary`
            The multipart's headers.
        data : `bytes`, `bytes-like`
            The multipart's data.
        
        Raises
        ------
        EOFError
            Connection lost before enough data was received.
        PayloadError
            Invalid data received.
        CancelledError
            If the reader task is cancelled not by receiving eof.
        ContentEncodingError
            - `'content_encoding'` was given as `'br'` meanwhile brotli or brotlipy are not installed.
            - `'content_encoding'` is not an from the expected values.
        StopAsyncIteration
            The payload contains no more fields.
        """
        content_type_header = headers.get(CONTENT_TYPE, None)
        if content_type_header is None:
            boundary = None
        else:
            content_type, content_type_parsing_error = parse_content_type(content_type_header)
            boundary = content_type.get_parameter('boundary', '')
            if not boundary:
                boundary = None
        
        if (boundary is None):
            raise PayloadError(f'Missing boundary: {boundary}.')
        
        is_first = True
        while True:
            is_more, headers, data = await self._read_single_multipart(boundary, is_first)
            if (data is not None):
                try:
                    transfer_encoding = headers[CONTENT_TRANSFER_ENCODING]
                except KeyError:
                    pass
                else:
                    transfer_encoding = transfer_encoding.casefold()
                    if transfer_encoding == 'base64':
                        data = base64_decode(data)
                    elif transfer_encoding == 'quoted-printable':
                        data = qp_decode(data)
                    elif transfer_encoding in ('binary', '8bit', '7bit'):
                        pass
                    else:
                        raise PayloadError(f'Unknown transfer encoding: {transfer_encoding!r}.')
                
                try:
                    content_encoding = headers[CONTENT_ENCODING]
                except KeyError:
                    pass
                else:
                    content_encoding = content_encoding.casefold()
                    decompressor = get_decompressor_for(content_encoding)
                    if (decompressor is not None):
                        try:
                            data = decompressor.decompress(data)
                        except COMPRESSION_ERRORS as exception:
                            raise PayloadError('Cannot decompress data.') from exception
                
                yield headers, data
            
            if not is_more:
                return
            
            is_first = False
            continue
    
    
    async def read_web_socket_frame(self, client_side, max_size):
        """
        Reads a web socket frame.
        
        This method is a coroutine.
        
        Returns
        -------
        frame : ``WebSocketFrame``
            The read web socket frame.
        
        Raises
        ------
        EofError
            Connection lost before a full frame was received.
        PayloadError
            Payload length over max size limit.
        CancelledError
            If the reader task is cancelled not by receiving eof.
        """
        head_0, head_1 = await self.read_exactly(2)
        
        if ((head_1 & 0b10000000) >> 7) == client_side:
            raise WebSocketProtocolError('Incorrect masking.')
        
        length = head_1 & 0b01111111
        
        if length == 126:
            data = await self.read_exactly(2)
            length, = UNPACK_LENGTH_2(data)
        elif length == 127:
            data = await self.read_exactly(8)
            length, = UNPACK_LENGTH_3(data)
        
        if (max_size is not None) and length > max_size:
            raise PayloadError(f'Payload length exceeds size limit ({length} > {max_size} bytes).')
        
        # Read the data.
        if client_side:
            data = await self.read_exactly(length)
        else:
            mask = await self.read_exactly(4)
            data = await self.read_exactly(length)
            data = apply_web_socket_mask(mask, data)
        
        return WebSocketFrame._from_fields(head_0, data)
    
    
    def get_payload_reader_task(self, message):
        """
        Gets payload reader task for the given raw http message.
        
        Parameters
        ----------
        message : ``RawMessage``
            Raw http message. Can be either request or response.
        
        Returns
        -------
        payload_reader_task : `None`, `GeneratorType`
            Payload reader task if applicable.
        """
        length = message.headers.get(CONTENT_LENGTH, None)
        if (length is not None):
            if length.isdigit():
                length = int(length)
            else:
                raise PayloadError(f'{CONTENT_LENGTH} must be a non negative int, got: {length!r}.')
        
        if (not message.upgraded):
            if message.status == 204:
                return None
            
            if message.chunked:
                decompressor = get_decompressor_for(message.encoding)
                if decompressor is None:
                    return self._read_chunked
                else:
                    return partial_func(self._read_chunked_encoded, decompressor)
            
            if (length is not None) and (length > 0):
                decompressor = get_decompressor_for(message.encoding)
                if decompressor is None:
                    return partial_func(self._read_exactly, length, False)
                else:
                    return partial_func(self._read_exactly_encoded, length, decompressor)
        
        if isinstance(message, RawRequestMessage) and (message.method == METHOD_CONNECT):
            message.upgraded = True
            return self._read_until_eof
        
        if isinstance(message, RawResponseMessage) and message.status >= 199 and (length is None):
            if message.status == 204:
                return None
            
            if message.chunked:
                decompressor = get_decompressor_for(message.encoding)
                if decompressor is None:
                    return self._read_chunked
                else:
                    return partial_func(self._read_chunked_encoded, decompressor)
            
            return self._read_until_eof
        
        return None
    
    
    async def _read_chunked_by_chunk(self):
        """
        Reads a chunked http message by chunk and yields it back. Used by other payload readers.
        
        This method is a coroutine generator.
        
        Yields
        -------
        chunk : `bytes | memoryview`
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        PayloadError
            - Chunk size is not hexadecimal.
            - Chunk not ends with CRLF.
        """
        while True:
            line = await self._read_until_as_one(b'\r\n')
            # strip chunk extension
            index = line.find(b';')
            if index != -1:
                line = line[:index]
            
            try:
                length = int(line, 16)
            except ValueError:
                raise PayloadError(
                    f'Not hexadecimal chunk size: {line!r}.'
                ) from None
            
            if length == 0:
                end = await self._read_exactly_as_one(2)
                if end != b'\r\n':
                    raise PayloadError(
                        f'Received chunk does not end with b\'\\r\\n\', instead with: {end}.'
                    )
                break
            
            async for chunk in self._read_exactly_by_chunk(length):
                yield chunk
            
            end = await self._read_exactly_as_one(2)
            if end != b'\r\n':
                raise PayloadError(
                    f'Received chunk does not end with b\'\\r\\n\', instead with: {end}.'
                )
    
    
    async def _read_chunked(self, payload_stream):
        """
        Payload reader task, that reads a chunked http message, till it ends.
        
        This method is a coroutine.
        
        Parameters
        ----------
        payload_stream : ``PayloadStream``
            Payload buffer to read into.
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        PayloadError
            - Chunk size is not hexadecimal.
            - Chunk not ends with CRLF.
        """
        async for chunk in self._read_chunked_by_chunk():
            payload_stream.add_received_chunk(chunk)
    
    
    async def _read_exactly_encoded(self, length, decompressor, payload_stream):
        """
        Payload reader task, that reads exactly `n` bytes from the protocol, then decompresses it with the given
        decompress object.
        
        This method is a coroutine.
        
        Parameters
        ----------
        length : `int`
            The amount of bytes to read.
        
        decompressor : `ZLIB_DECOMPRESSOR`, `BROTLI_DECOMPRESSOR`
            Decompressor used to decompress the data after receiving it.
        
        payload_stream : ``PayloadStream``
            Payload buffer to read into.
        
        Returns
        -------
        collected : `bytes`
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        PayloadError
            Cannot decompress a chunk.
        """
        async for chunk in self._read_exactly_by_chunk(length):
            try:
                chunk = decompressor.decompress(chunk)
            except COMPRESSION_ERRORS:
                raise PayloadError('Cannot decompress chunk.') from None
            
            payload_stream.add_received_chunk(chunk)
        
        
        # Flush decompressor at the end in case it is required.
        try:
            chunk = decompressor.flush()
        except COMPRESSION_ERRORS:
            raise PayloadError('Cannot decompress chunk.') from None
        
        if chunk:
            payload_stream.add_received_chunk(chunk)
    
    
    async def _read_chunked_encoded(self, decompressor, payload_stream):
        """
        Payload reader task, that reads a chunked http message, till it ends. After receiving a "chunk", decompresses
        it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        decompressor : `ZLIB_DECOMPRESSOR`, `BROTLI_DECOMPRESSOR`
            Decompressor used to decompress the data after receiving it.
        
        payload_stream : ``PayloadStream``
            Payload buffer to read into.
            
        Returns
        -------
        collected : `bytes`
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        PayloadError
            - Chunk size is not hexadecimal.
            - Chunk not ends with CRLF.
            - Cannot decompress a chunk.
        """
        async for chunk in self._read_chunked_by_chunk():
            try:
                chunk = decompressor.decompress(chunk)
            except COMPRESSION_ERRORS:
                raise PayloadError('Cannot decompress chunk.') from None
            
            payload_stream.add_received_chunk(chunk)
        
        
        # Flush decompressor at the end in case it is required.
        try:
            chunk = decompressor.flush()
        except COMPRESSION_ERRORS:
            raise PayloadError('Cannot decompressor chunk.') from None
        
        if chunk:
            payload_stream.add_received_chunk(chunk)
    
    
    def isekai_into(self, other_class):
        """
        Copies the protocol's attributes to an other one.
        
        Can be used to change a protocol's type.
        
        Parameters
        ----------
        other_class : `type`
            The other protocol asynchronous protocol type to isekai self into.
        
        Returns
        -------
        new : `other_class
        """
        new = object.__new__(other_class)
        
        new._loop = self._loop
        new._transport = self._transport
        new._exception = self._exception
        new._chunks = self._chunks
        new._offset = self._offset
        new._at_eof = self._at_eof
        new._payload_reader = self._payload_reader
        new._payload_stream = self._payload_stream
        new._paused = self._paused
        
        return new
    

class HttpReadWriteProtocol(ReadWriteProtocolBase, HttpReadProtocol):
    """
    Asynchronous read-write protocol implementation. Stuffed full of http read methods.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    _chunks : `deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_stream : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``.payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    _transport : `None`, `object`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    _drain_waiter : `None`, ``Future``
        A future, what is used to block the writing task, till it's writen data is drained.
    """
    __slots__ = ()
    
    def write_http_request(self, method, path, headers, version = HttpVersion11):
        """
        Writes an http request to the protocol's transport.
        
        Parameters
        ----------
        method : `str`
            The request's method.
        path : `str`
            The request's path.
        headers : ``IgnoreCaseMultiValueDictionary``
            Request headers.
        version : ``HttpVersion`` = `HttpVersion11`, Optional
            Http version of the request. Defaults to `HttpVersion11`.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        transport = self._transport
        if transport is None:
            raise RuntimeError(f'Protocol has no attached transport; self = {self!r}.')
        
        result = [f'{method} {path} HTTP/{version.major}.{version.major}\r\n']
        extend = result.extend
        for k, v in headers.items():
            extend((k, ': ', v, '\r\n'))
        
        result.append('\r\n')
        transport.write(''.join(result).encode())
    
    
    def write_http_response(self, status, headers, version = HttpVersion11, body = None):
        """
        Writes an http response to the protocol's transport.
        
        Parameters
        ----------
        status : `http.HTTPStatus`
            The response's status.
        headers : ``IgnoreCaseMultiValueDictionary``
            Response headers.
        version : ``HttpVersion`` = `HttpVersion11`, Optional
            Http version of the response. Defaults to `HttpVersion11`.
        body : `None`, `bytes-like` = `None`, Optional
            Http response body.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        transport = self._transport
        if transport is None:
            raise RuntimeError(f'Protocol has no attached transport; self = {self!r}.')
        
        result = [
            'HTTP/',
            str(version.major),
            '.',
            str(version.minor),
            ' ',
            str(status.value),
            ' ',
            str(status.phrase),
            '\r\n',
        ]
        
        for header_key, header_value in headers.items():
            result.extend((header_key, ': ', header_value, '\r\n'))
        
        result.append('\r\n')
        
        transport.write(''.join(result).encode())
        if (body is not None) and body:
            transport.write(body)
    
    
    def write_web_socket_frame(self, frame, client_side):
        """
        Writes a web socket frame to the protocol.
        
        Parameters
        ----------
        frame : ``WebSocketFrame``
            The web socket frame to write.
        client_side : `bool`
            Whether the respective web socket is client or server side.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        transport = self._transport
        if transport is None:
            raise RuntimeError(f'Protocol has no attached transport; self = {self!r}.')
        
        # Prepare the header.
        head_0 = frame.head_0
        head_1 = client_side << 7
        
        length = len(frame.data)
        if length < 126:
            header = PACK_LENGTH_1(head_0, head_1 | length)
        elif length < 65536:
            header = PACK_LENGTH_2(head_0, head_1 | 126, length)
        else:
            header = PACK_LENGTH_3(head_0, head_1 | 127, length)
        transport.write(header)
        
        # prepare the data.
        if client_side:
            mask = getrandbits(32).to_bytes(4, 'big')
            transport.write(mask)
            data = apply_web_socket_mask(mask, frame.data,)
        else:
            data = frame.data
        
        transport.write(data)

    
    @copy_docs(HttpReadProtocol.isekai_into)
    def isekai_into(self, other_class):
        new = HttpReadProtocol.isekai_into(self, other_class)
        
        new._drain_waiter = self._drain_waiter
        
        return new



__all__ = ()

from ...utils import copy_docs

from ..quoting import quote, unquote

from .constants import URL_PART_FLAG_DECODED_SET, URL_PART_FLAG_ENCODED_SET
from .url_part_base import URLPartBase


class URLPartBaseParsed(URLPartBase):
    """
    Represents an url's part that can be parsed.
    
    Attributes
    ----------
    _decoded : `None | str`
        Decoded value.
    
    _encoded : `None | str`
        Encoded value.
    
    _flags : `int`
        Bitwise flags representing how the object is constructed.
    
    parsed : `None | object`
        The parsed value.
    """
    __slots__ = ('parsed',)
    
    def __new__(cls):
        """
        Use either ``.create_from_decoded``, ``.create_from_encoded`` or ``.create_from_parsed`` instead.
        
        Raises
        ------
        RuntimeError
        """
        raise RuntimeError(
            'Use either `.create_from_decoded`, `.create_from_encoded` or `.create_from_parsed` instead.'
        )
    
    
    def __eq__(self, other):
        """Returns whether the two url parts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # parsed
        if self.parsed != other.parsed:
            return False
        
        return True
    
    
    def _parse(value, encoded):
        """
        Parses the given value.
        
        Parameters
        ----------
        value : `None | str`
            The value to parse.
        
        encoded : `bool`
            Whether an encoded value is passed.
        
        Returns
        -------
        value : `None | object`
        """
        if encoded:
            value = unquote(value)
        
        return value
    
    
    @staticmethod
    def _serialize(value, encoded):
        """
        Serializes the given value.
        
        Parameters
        ----------
        value : `None | object`
            The value serialize.
        
        encoded : `bool`
            Whether an encoded value is requested.
        
        Returns
        -------
        value : `None | str`
        """
        if encoded:
            value = quote(value)
        
        return value
    
    
    @classmethod
    @copy_docs(URLPartBase.create_from_decoded)
    def create_from_decoded(cls, decoded):
        parsed = cls._parse(decoded, False)
        
        self = object.__new__(cls)
        self._decoded = None
        self._encoded = None
        self._flags = 0
        self.parsed = parsed
        return self
    
    
    @classmethod
    @copy_docs(URLPartBase.create_from_encoded)
    def create_from_encoded(cls, encoded):
        parsed = cls._parse(encoded, True)
        
        self = object.__new__(cls)
        self._decoded = None
        self._encoded = None
        self._flags = 0
        self.parsed = parsed
        return self


    @classmethod
    @copy_docs(URLPartBase.create_from_encoded)
    def create_from_parsed(cls, parsed):
        """
        Creates a new url part from a parsed value.
        
        Parameters
        ----------
        parsed : `None | object`
            Parsed value.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._decoded = None
        self._encoded = None
        self._flags = 0
        self.parsed = parsed
        return self
    
    
    @property
    @copy_docs(URLPartBase.decoded)
    def decoded(self):
        flags = self._flags
        if flags & URL_PART_FLAG_DECODED_SET:
            decoded = self._decoded
        
        else:
            decoded = self._serialize(self.parsed, False)
            
            self._decoded = decoded
            self._flags = flags | URL_PART_FLAG_DECODED_SET
        
        return decoded
    
    
    @property
    @copy_docs(URLPartBase.encoded)
    def encoded(self):
        flags = self._flags
        if flags & URL_PART_FLAG_ENCODED_SET:
            encoded = self._encoded
        
        else:
            encoded = self._serialize(self.parsed, True)
            
            self._encoded = encoded
            self._flags = flags | URL_PART_FLAG_ENCODED_SET
        
        return encoded



__all__ = ()

from re import compile as re_compile
from sys import platform, version_info

from .word_pattern import create_word_pattern


COMPLEX_RP = re_compile(
    '((?:\\d(?:_?\\d)*\\.\\d(?:_?\\d)*|\\d(?:_?\\d)*\\.|\\.\\d(?:_?\\d)*)(?:[eE][+-]?\\d(?:_?\\d)*)?[jJ])'
)
FLOAT_RP = re_compile(
    '((?:\\d(?:_?\\d)*\\.\\d(?:_?\\d)*|\\d(?:_?\\d)*\\.|\\.\\d(?:_?\\d)*)(?:[eE][+-]?\\d(?:_?\\d)*)?)'
)
INTEGER_HEXADECIMAL_RP = re_compile('(0[xX](?:_?[0-9a-fA-F])+)')
INTEGER_DECIMAL_RP = re_compile('(\\d(?:_?\\d)*)')
INTEGER_OCTAL_RP = re_compile('(0[oO](?:_?[0-7])+)')
INTEGER_BINARY_RP = re_compile('(0[bB](?:_?[01])+)')
IDENTIFIER_RP = re_compile('([a-zA-Z_][a-zA-Z_0-9]*)')

ATTRIBUTE_ACCESS_OPERATOR = '.'
KEYWORD_ELLIPSIS = '...'
ESCAPE = '\\'

BUILTIN_CONSTANTS = {
    Ellipsis, False, None, NotImplemented, True
}

BUILTIN_CONSTANT_NAMES = {
    KEYWORD_ELLIPSIS,
    *(repr(value) for value in BUILTIN_CONSTANTS),
}

KEYWORDS = {
    'as', 'assert', 'async', 'await', 'break', 'class', 'continue', 'def', 'del', 'elif', 'else', 'except', 'finally',
    'for', 'from', 'global', 'if', 'import', 'lambda', 'nonlocal', 'pass', 'raise', 'return', 'try', 'while', 'with',
    'yield'
}

BUILTIN_VARIABLES = {
    __import__, abs, all, any, ascii, bin, bool, bytearray, bytes, chr, classmethod, compile, complex, delattr, dict,
    dir, divmod, enumerate, eval, filter, float, format, frozenset, getattr, globals, hasattr, hash, hex, id, input,
    int, isinstance, issubclass, iter, len, list, locals, map, max, memoryview, min, next, object, oct, open, ord, pow,
    print, property, range, repr, reversed, round, set, setattr, slice, sorted, staticmethod, str, sum, super, tuple,
    type, vars, zip,
    *((aiter, anext) if version_info >= (3, 10) else ()),
}


BUILTIN_VARIABLE_NAMES = {
    value.__name__ for value in BUILTIN_VARIABLES
}

BUILTIN_EXCEPTIONS = {
    ArithmeticError, AssertionError, AttributeError, BaseException, BlockingIOError, BrokenPipeError, BufferError,
    BytesWarning, ChildProcessError, ConnectionAbortedError, ConnectionError, ConnectionRefusedError,
    ConnectionResetError, DeprecationWarning, EOFError, EnvironmentError, Exception, FileExistsError,
    FileNotFoundError, FloatingPointError, FutureWarning, GeneratorExit, IOError, ImportError, ImportWarning,
    IndentationError, IndexError, InterruptedError, IsADirectoryError, KeyError, KeyboardInterrupt, LookupError,
    MemoryError, NameError, NotADirectoryError, NotImplementedError, OSError, OverflowError, PendingDeprecationWarning,
    PermissionError, ProcessLookupError, RecursionError, ReferenceError, ResourceWarning, RuntimeError, RuntimeWarning,
    StopAsyncIteration, StopIteration, SyntaxError, SyntaxWarning, SystemError, SystemExit, TabError, TimeoutError,
    TypeError, UnboundLocalError, UnicodeDecodeError, UnicodeEncodeError, UnicodeError, UnicodeTranslateError,
    UnicodeWarning, UserWarning, ValueError, Warning, ZeroDivisionError,
    *((WindowsError,) if platform == 'win32' else ()),
    *((ModuleNotFoundError,) if version_info >= (3, 6) else ()),
    *((EncodingWarning,) if version_info >= (3, 10) else ()),
    *((BaseExceptionGroup, ExceptionGroup,) if version_info >= (3, 11) else ()),
    *((PythonFinalizationError,) if version_info >= (3, 13) else ()),
}

BUILTIN_EXCEPTION_NAMES = {
    value.__name__ for value in BUILTIN_EXCEPTIONS
}

MAGIC_FUNCTION_NAMES = {
    '__abs__', '__add__', '__aenter__', '__aexit__', '__aiter__', '__and__', '__anext__', '__await__', '__bool__',
    '__bytes__', '__call__', '__complex__', '__contains__', '__del__', '__delattr__', '__delete__', '__delitem__',
    '__dir__', '__divmod__', '__enter__', '__eq__', '__exit__', '__float__', '__floordiv__', '__format__', '__ge__',
    '__get__', '__getattr__', '__getattribute__', '__getitem__', '__gt__', '__hash__', '__iadd__', '__iand__',
    '__ifloordiv__', '__ilshift__', '__imatmul__', '__imod__', '__imul__', '__index__', '__init__', '__instancecheck__',
    '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__', '__isub__', '__iter__', '__itruediv__', '__ixor__',
    '__le__', '__len__', '__length_hint__', '__lshift__', '__lt__', '__matmul__', '__missing__', '__mod__', '__mul__',
    '__ne__', '__neg__', '__new__', '__next__', '__or__', '__pos__', '__pow__', '__prepare__', '__radd__', '__rand__',
    '__rdivmod__', '__repr__', '__reversed__', '__rfloordiv__', '__rlshift__', '__rmatmul__', '__rmod__', '__rmul__',
    '__ror__', '__round__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', '__rtruediv__', '__rxor__', '__set__',
    '__setattr__', '__setitem__', '__str__', '__sub__', '__subclasscheck__', '__truediv__', '__xor__'
}

MAGIC_VARIABLE_NAMES = {
    '__annotations__', '__bases__', '__class__', '__closure__', '__code__', '__defaults__', '__dict__', '__doc__',
    '__file__', '__func__', '__globals__', '__kwdefaults__', '__module__', '__mro__', '__name__', '__objclass__',
    '__qualname__', '__self__', '__slots__', '__weakref__'
}

PUNCTUATIONS = {
    '(', ')', ',', ':', ';', '[', ']', '{', '}'
}

OPERATOR_WORDS = {
    'and', 'in', 'is', 'not', 'or'
}

OPERATORS = {
    '!=', '%', '%=', '&', '&=', '*', '**', '**=', '*=', '+', '+=', '-', '-=', '->', '.', '...', '/', '//', '//=', '/=',
    '<', '<<', '<<=', '<=', '=', '==', '>', '>=', '>>', '>>=', '@', '@=', '\\', '^', '^=', '|', '|='
}

STRING_STARTER_RP = re_compile('(r[fb]?|[fb]r?|b|f)?(\'{3}|\"{3}|\'|\")')
STRING_END_SINGLE_RP = re_compile('(.*?)(?<!\\\\)\'|\'')
STRING_END_DOUBLE_RP = re_compile('(.*?)(?<!\\\\)\"|\'')
STRING_MULTI_LINE_END_SINGLE_RP = re_compile('(.*?)(?<!\\\\)\'\'\'|\'\'\'')
STRING_MULTI_LINE_END_DOUBLE_RP = re_compile('(.*?)(?<!\\\\)\"\"\"|\"\"\"')

SPACE_MATCH_RP = re_compile('([ \t]+)')

FORMAT_STRING_MATCH_STRING = re_compile('(.*?)(\\{\\{?|\\}\\}?|\n)')

CONSOLE_PREFIX_RP = re_compile('(>>>>?|\\.\\.\\.\\.?| *\\.\\.\\.\\:|In \\[\\d+\\]\\:)( [ \t]*)')

FORMAT_STRING_POSTFIX_RP = re_compile('(![sraSRA])\\}')

PUNCTUATION_WP = create_word_pattern(PUNCTUATIONS)
OPERATOR_WP = create_word_pattern(OPERATORS)



__all__ = ('AttributeError', 'AttributeErrorBase', 'RichAttributeErrorBaseType',)


ATTRIBUTE_ERROR_HAS_RICH_SLOTS = hasattr(AttributeError, 'obj') and hasattr(AttributeError, 'name')
AttributeErrorBase = AttributeError


class AttributeError(AttributeError):
    """
    Represents a rich attribute error.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute's name that was not found.
    instance : `object`
        The instance that does not have the respective attribute.
    """
    if ATTRIBUTE_ERROR_HAS_RICH_SLOTS:
        __slots__ = ()
        
        attribute_name = AttributeErrorBase.name
        instance = AttributeErrorBase.obj
    else:
        __slots__ = ('attribute_name', 'instance')
    
    def __new__(cls, instance, attribute_name):
        self = AttributeErrorBase.__new__(cls, instance, attribute_name)
        self.instance = instance
        self.attribute_name = attribute_name
        return self
    
    __init__ = object.__init__
    
    
    def __repr__(self):
        """Returns the attribute error's representations."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' instance = ')
        repr_parts.append(repr(self.instance))
        
        repr_parts.append(', attribute_name = ')
        repr_parts.append(repr(self.attribute_name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


class RichAttributeErrorBaseType:
    """
    Base type for generating rich attribute error messages.
    """
    __slots__ = ()
    
    def __getattr__(self, attribute_name):
        """Drops a rich attribute error."""
        raise AttributeError(self, attribute_name)



__all__ = ()

from types import FunctionType

SlotWrapperType = type(object.__lt__)
BuiltinFunctionOrMethodType = type(object.__new__)


def _get_type_name(type_):
    """
    Gets the type's name.
    
    Parameters
    ----------
    type_ : `type`
        The type to get its name of.
    
    Returns
    -------
    type_name : `str`
    """
    # This ignores ignores user level modifications.
    type_name = type.__dict__['__name__'].__get__(type_, type(type_))
    
    # Note that you can only set `str` as `.__name__`, but subclasses are also allowed.
    if type(type_name) is not str:
        type_name = str(type_name)
    
    return type_name


def _get_value_representation(value):
    """
    Gets the representation of the given value.
    
    Parameters
    ----------
    value : `object`
        The value to get its representation of.
    
    Returns
    -------
    representation : `str`
    """
    try:
        representation = repr(value)
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        representation = f'<{_get_type_name(type(value))!s}>'
    
    else:
        # `repr()` can only return `str` instance including sub-types, so no need for `instance` check.
        if type(representation) is not str:
            representation = str(representation)
    
    return representation


def _get_exception_parameters(exception):
    """
    Returns the parameters the exception was created with.
    
    Parameters
    ----------
    exception : ``BaseException``
        Exception to get its parameters of.
    
    Returns
    -------
    exception_parameters : `None | tuple<object>`
    """
    # Exceptions always have `args` set.
    # Directly attach the field descriptor instead of the attribute to avoid user level modifications.
    exception_parameters = BaseException.args.__get__(exception, type(exception))
    if type(exception_parameters) is tuple:
        return exception_parameters
    
    # Do not use `instance` check if `__class__` is overwritten. Later may add mor conditions here depending how hard
    # people mess it up.
    exception_parameters_type = type(exception_parameters)
    if (
        '__class__' not in
        type.__dict__['__dict__'].__get__(exception_parameters_type, type(exception_parameters_type)).keys()
    ):
        if isinstance(exception_parameters, tuple):
            return tuple(exception_parameters)
    
    return None


def _are_exception_constructors_default(exception):
    """
    Gets whether the exception's constructor(s) are the default ones.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception instance to check.
    
    Returns
    -------
    are_default : `bool`
    """
    exception_type = type(exception)
    
    exception_type_new = type.__getattribute__(exception_type, '__new__')
    exception_type_init = type.__getattribute__(exception_type, '__init__')
    
    # is it pypy?
    if FunctionType is SlotWrapperType:
        if (exception_type_new is not BaseException.__new__) and (exception_type_new is not object.__new__):
            return False
        
        if (exception_type_init is not BaseException.__init__) and (exception_type_init is not object.__init__):
            return False
        
        return True
    
    # Not pypy
    if type(exception_type_new) is not BuiltinFunctionOrMethodType:
        return False
    
    if type(exception_type_init) is not SlotWrapperType:
        return False
    
    return True


def _get_exception_representation_fallback(exception):
    """
    Gets exception representation by building it.
    
    Parameters
    ----------
    exception : ``BaseException``
        The respective exception instance.
    
    Returns
    -------
    raw_exception_representation : `str`
    """
    parts = ['> repr(exception) raised, using fallback representation.\n']
    
    parts.append(_get_type_name(type(exception)))
    parts.append('(')
    
    exception_parameters = _get_exception_parameters(exception)
    if (exception_parameters is not None):
        length = len(exception_parameters)
        if length:
            index = 0
            while True:
                element = exception_parameters[index]
                parts.append(_get_value_representation(element))
                
                index += 1
                if index == length:
                    break
                
                parts.append(', ')
                continue
    
    parts.append(')')
    return ''.join(parts)


def _get_exception_representation_simple(exception):
    """
    Tries to get simple exception representation.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to get it's representation of.
    
    Returns
    -------
    exception_representation : `None`, `str`
        Returns `None` if simple representation is not available.
    """
    if not _are_exception_constructors_default(exception):
        return None
    
    exception_parameters = _get_exception_parameters(exception)
    if (exception_parameters is None) or len(exception_parameters) > 1:
        return None
    
    exception_representation = _get_type_name(type(exception))
    if len(exception_parameters) == 1:
        # Get value representation only if value is not `str`.
        value = exception_parameters[0]
        if issubclass(type(value), str):
            value = str(value)
        else:
            value = _get_value_representation(value)
        
        exception_representation = f'{exception_representation!s}: {value!s}'
    
    return exception_representation


def get_exception_representation_generic(exception):
    """
    Gets the exception's representation.
    
    Parameters
    ----------
    exception : ``BaseException``
        The respective exception instance.
    
    Returns
    -------
    exception_representation : `str`
    """
    exception_representation = _get_exception_representation_simple(exception)
    if exception_representation is not None:
        return exception_representation
    
    try:
        exception_representation = repr(exception)
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        pass
    else:
        if type(exception_representation) is not str:
            exception_representation = str(exception_representation)
        return exception_representation
    
    return _get_exception_representation_fallback(exception)



__all__ = ()


def is_exception(exception):
    """
    Returns whether the given value is an exception.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to check.
    
    Returns
    -------
    is_exception : `bool`
    """
    return issubclass(type(exception), BaseException)
